import random
import string
import uuid
from dataclasses import dataclass
from typing import Any

import pydash
from adafri.utils.queries import (
    cloud_firestore_query_builder,
    cloud_firestore_query_formatter,
)
from adafri.v1 import version
from adafri.v1.account.models.account import Account
from adafri.v1.files import IMAGES_EXTENSIONS, File
from adafri.v1.files.utils import upload_image_from_url
from adafri.v1.gads.data import DISPLAY_ADS_REQUIREMENTS
from adafri.v1.gads.models.client_customer_id import ClientCustomerId
from firebase_admin.firestore import firestore

from .....utils import (
    ArrayUtils,
    BaseClass,
    DictUtils,
    RequestFields,
    get_object_model_class,
    get_request_fields,
    init_class_kwargs,
    integer,
)
from .....utils.response import ApiResponse, Error, ResponseStatus, StatusCode
from ...models.adgroups.ads.base import BaseAd, BaseAdFieldsProps
from .ad_fields import (
    DISPLAY_ADS_COLLECTION,
    IMAGE_AD_PICKABLE_FIELDS,
    ImageAdFields,
    ImageAdFieldsProps,
)

DISPLAY_ADS_UPLOAD_COLLECTION = "image_upload"

def filter_image_ad_request_fields(fields, default_fields = IMAGE_AD_PICKABLE_FIELDS):
    request_fields = get_request_fields(fields, default_fields, [default_fields[0]])
    if request_fields is None or bool(request_fields) is False:
        request_fields = [default_fields[0]]
    return request_fields


@dataclass(init=False)
class ImageAdToPublish(BaseClass):
    id: str
    name: str
    mediaId: int
    assetId: int
    hash: str
    width: int
    height: int
    urls: str
    url: str
    previewUrl: str
    resourceName: str
    sourceUrl: str
    type: str
    mimeType: str
    referenceId: str
    owner: str
    fileSize: str
    clientCustomerId: str
    def __init__(self, ad=None, **kwargs):
        (cls_object, keys, data_args) = init_class_kwargs(self, ad, None, None, None, [], **kwargs)
        for key in keys:
            if key == ImageAdFields.fileSize:
                if bool(cls_object[key]) is True:
                    setattr(self, key, integer(cls_object[key]))
                else:
                    setattr(self, key, None)
            elif key in [ImageAdFields.mediaId, ImageAdFields.assetId]:
                if bool(cls_object[key]) is True:
                    setattr(self, key, integer(cls_object[key]))
                else:
                    setattr(self, key, None)
            elif key == ImageAdFields.width:
                if bool(cls_object[key]) is True:
                    setattr(self, key, integer(cls_object[key]))
                else:
                    setattr(self, key, None)
            elif key == ImageAdFields.height:
                if bool(cls_object[key]) is True:
                    setattr(self, key, integer(cls_object[key]))
                else:
                    setattr(self, key, None)
            else:
                setattr(self, key, cls_object[key]) 
        if bool(self.urls) and bool(self.url) is False:
            self.url = self.urls
        if bool(self.mediaId) and bool(self.assetId) is False:
            self.assetId = self.mediaId
    def to_dict(self, field=None):
        return DictUtils.remove_none_values({
            "id": self.id,
            "name": self.name,
            "mediaId": self.mediaId,
            "assetId": self.assetId,
            "width": self.width,
            "height": self.height,
            "urls": self.urls,
            "url": self.url,
            "previewUrl": self.previewUrl,
            "resourceName": self.resourceName,
            "sourceUrl": self.sourceUrl,
            "type": self.type,
            "hash": self.hash,
            "mimeType": self.mimeType,
            "owner": self.owner,
            "fileSize": self.fileSize,
            "referenceId": self.referenceId,
            "clientCustomerId": self.clientCustomerId
        })
    @staticmethod
    def from_dict(_obj: Any) -> 'ImageAdToPublish':
        obj = {}
        if bool(_obj) is True:
            if isinstance(_obj, dict):
                obj = _obj.copy()
                if 'urls' in obj and 'url' not in obj:
                    obj['url'] = obj['urls']
                if 'mediaId' in obj and 'assetId' not in obj:
                    obj['assetId'] = obj['mediaId']
            elif isinstance(_obj, ImageAdToPublish):
                obj = _obj.to_dict()
                if 'urls' in obj and 'url' not in obj:
                    obj['url'] = obj['urls']
                if 'mediaId' in obj and 'assetId' not in obj:
                    obj['assetId'] = obj['mediaId']
            elif isinstance(_obj, (str, int)):
                obj = {"id": str(_obj)} 
            else:
                obj = _obj
        cls_object, keys = get_object_model_class(obj, ImageAdToPublish, None);
        return ImageAdToPublish(cls_object)
    
    @staticmethod
    def convert_to_image_ad(value):
        ad = ImageAd();
        if bool(value.urls) is True:
            ad.urls = value.urls
            ad.url_image = value.urls
        if bool(value.name) is True:
            ad.name = value.name
        if bool(value.mediaId) is True:
            ad.mediaId = value.mediaId
            ad.assetId = value.mediaId
        if bool(value.type) is True:
            ad.type = value.type
        if bool(value.width) is True:
            ad.width = value.width
        if bool(value.height) is True:
            ad.height = value.height
        if bool(value.previewUrl) is True:
            ad.previewUrl = value.previewUrl
        if bool(value.sourceUrl) is True:
            ad.sourceUrl = value.sourceUrl
        if bool(value.mimeType) is True:
            ad.mimeType = value.mimeType
        if bool(value.owner) is True:
            ad.owner = value.owner
        if bool(value.fileSize) is True:
            ad.fileSize = value.fileSize
        if bool(value.clientCustomerId) is True:
            ad.clientCustomerId = value.clientCustomerId
        if bool(value.referenceId) is True:
            ad.referenceId = value.referenceId
        return ad
    
    def to_image_ad(self, owner = None):
        if bool(owner) is True:
            self.owner = owner
        return ImageAdToPublish.convert_to_image_ad(self)
@dataclass(init=False)
class ImageAd(BaseAd):
    mediaId: int
    assetId: int
    mimeType: str
    fileSize: int
    hash: str
    previewUrl: str
    referenceId: str | int
    type: str
    sourceUrl: str
    urls: str
    url: str
    url_image: str
    usedForRectangle: bool
    usedForSquare: bool
    width: int
    height: int
    ad_name: str
    ad_type: str
    state: str
    displayUrl: str
    finalAppUrls: list[str]
    finalMobileUrls: list[str]
    finalUrls: list[str]
    google_ads_image_ad: dict
    __baseFields = ImageAdFieldsProps
    class_object = None;
    adgroup: Any | None = None
    customer: ClientCustomerId | None = None
    account: Account | None = None

    def __init__(self, ad=None, **kwargs):
        _ad = ad
        if type(ad) is str:
            _ad = {"id": ad}
        (cls_object, keys, data_args) = init_class_kwargs(self, _ad, IMAGE_AD_PICKABLE_FIELDS, ImageAdFieldsProps, DISPLAY_ADS_COLLECTION, ['id'], **kwargs)
        super().__init__(**{**data_args, "ad": _ad});
        for key in keys:
            if BaseAdFieldsProps.get(key) is None:
                if key == ImageAdFields.account:
                    setattr(self, key, Account.from_dict(cls_object[key]))
                elif key == ImageAdFields.customer:
                    setattr(self, key, ClientCustomerId.from_dict(cls_object[key]))
                elif key == ImageAdFields.adgroup:
                    setattr(self, key, cls_object[key])
                elif key == ImageAdFields.fileSize:
                    setattr(self, key, integer(cls_object[key], None, True))
                else:
                    setattr(self, key, cls_object[key])
        if bool(self.url_image) is False:
            if bool(self.urls) is True:
                self.url_image = self.urls
        self.class_object = ImageAd


    @staticmethod
    def generate_model(_key_="default_value"):
        model = {};
        props = ImageAdFieldsProps
        for k in DictUtils.get_keys(props):
            if k == "api_version":
                model[k] = version
                continue;
            if _key_ in props[k] and props[k][_key_] is not None:
                model[k] = props[k][_key_];
        
        # Generate random campaign name
        random_suffix = ''.join(random.choices(string.ascii_uppercase + string.digits, k=6))
        model['name'] = f'Ad Display {random_suffix}'
        return model;

    @staticmethod
    def from_dict(ad: Any, db=None, collection_name=None) -> 'ImageAd':
        cls_object, keys = get_object_model_class(ad, ImageAd, ImageAdFieldsProps);
        _ad = ImageAd(cls_object, db=db, collection_name=collection_name)
        return _ad
    def generateAdgroup(self, campaign, data=None):
        if campaign is None:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error("Campaign not found","INVALID_REQUEST", 1))

        model = ImageAd.generate_model()
        model['accountId'] = campaign.accountId
        model['clientCustomerId'] = campaign.clientCustomerId
        if data is not None:
            model = DictUtils.merge_dict(data, model, True)
        # print('create ad with model', model)
        create = ImageAd().create(model, campaign)
        if create.status == 'error':
            return create
        # print('fields', CAMPAIGN_DISPLAY_PICKABLE_FIELDS)
        values = create.data
        # values = DictUtils.omit_fields(values, IMAGE_AD_PICKABLE_FIELDS + ['owner', 'id_campagne', 'ad_group_id'])
        # print('values', values)
        return ApiResponse(ResponseStatus.OK, StatusCode.status_200, values, None)
    
    @staticmethod
    def property_keys():
        keys = [ImageAdFields.url_image, 
                ImageAdFields.urls, 
                ImageAdFields.name, 
                ImageAdFields.ad_name, 
                ImageAdFields.ad_type,
                ImageAdFields.displayUrl,
                ImageAdFields.finalUrls,
                ImageAdFields.finalMobileUrls,
                ImageAdFields.finalAppUrls,
                ImageAdFields.fileSize,
                ImageAdFields.width,
                ImageAdFields.referenceId,
                ImageAdFields.mediaId,
                ImageAdFields.assetId,
                ImageAdFields.height]
        return keys
    @staticmethod
    def get_ad_group_props(value):
        if value is None:
            return {}
        return DictUtils.pick_fields(value, ImageAd.property_keys())
    def to_dict(self, fields=None):
        base = super().to_dict(fields)
        return DictUtils.remove_none_values({
            **base,
            "mediaId": self.mediaId,
            "assetId": self.assetId,
            "hash": self.hash,
            "mimeType": self.mimeType,
            "fileSize": self.fileSize,
            "previewUrl": self.previewUrl,
            "referenceId": self.referenceId,
            "type": self.type,
            "sourceUrl": self.sourceUrl,
            "urls": self.urls,
            "url": self.url,
            "url_image": self.url_image,
            "usedForRectangle": self.usedForRectangle,
            "usedForSquare": self.usedForSquare,
            "width": self.width,
            "height": self.height,
            "ad_name": self.ad_name,
            "ad_type": self.ad_type,
            "displayUrl": self.displayUrl,
            "finalAppUrls": self.finalAppUrls,
            "finalMobileUrls": self.finalMobileUrls,
            "finalUrls": self.finalUrls,
            "google_ads_image_ad": self.google_ads_image_ad,
            "customer": self.customer.to_dict().copy() if self.customer is not None else None,
            "account": self.account.to_dict().copy() if self.account is not None else None,
            "adgroup": self.adgroup,
            "state": self.state
    })
    @staticmethod
    def findAd(id, aacid=None):
        print('findAd', id, aacid)
        isID = False
        try:
            id = int(id)
        except Exception as e:
            isID = True
        if isID is True:
            ad = ImageAd.from_dict({"id": id}).get(IMAGE_AD_PICKABLE_FIELDS)
            if ad is None:
                return None
            return ad 
        if id is None or aacid is None or not isinstance(id, int) or not isinstance(aacid, str) or len(aacid) == 0:
            return None
        ad = ImageAd(None).query(ImageAd, "ad", query_params=[{"key": "accountId", "comp": "==", "value": aacid}, {"key": "ad_group_id", "comp": "==", "value": id}], first=True)
        return ad
    def getAds(self, adgroup, fields=['id', 'name'], response_type = 'dict'):
        # print('getAds', fields)
        isID = False
        if adgroup is None:
            return None
        adGroupId = adgroup.id
        try:
            adGroupId = int(adGroupId)
        except:
            isID = True
        if isID is False and adgroup.accountId is None:
            return None
        adgroups = []
        queryParams = [{"key": "accountId", "comp": "==", "value": adgroup.accountId}]
        if isID is False:
            queryParams.append({"key": "ad_group_id", "comp": "==", "value": adGroupId})
        else :
            queryParams.append({"key": "adgroup_id_parent", "comp": "==", "value": adGroupId})
        ads_ = self.query(ImageAd, "ad", query_params=queryParams)
        
        request_fields = filter_image_ad_request_fields(fields)
        for ad in ads_:
            if response_type == 'class':
                adgroups.append(ad)
            else:
                if fields is None:
                    adgroups.append(ad.to_dict())
                    continue
                adgroups.append(DictUtils.filter_by_fields(ad.to_dict(), request_fields))
        return adgroups
    
    def get(self, fields=None):
        if bool(self.id) is None:
            return None;
        request_fields = filter_image_ad_request_fields(fields)
        doc = self.document_reference().get();
        if doc.exists is False:
            return None;
        doc_dict = doc.to_dict()
        if 'id' in doc_dict and bool(doc_dict['id']) is False:
            del doc_dict['id']
        data = {"id": doc.id, **doc_dict}
        return ImageAd.from_dict(ad=ImageAd.from_dict(ad=data).to_dict(),  db=self.db, collection_name=self.collection_name);

    def getByUserId(self,id, fields=None):
        adgroups = [];
        request_fields = filter_image_ad_request_fields(fields)
        docs = self.collection().where('owner','==',id).get();
        for doc in docs:
            data = doc.to_dict();
            data['id'] = doc.id;
            adgroups.append(ImageAd.from_dict(ad=ImageAd.from_dict(ad=data).to_json(request_fields),  db=self.db, collection_name=self.collection_name));
        return adgroups;
    def normalize_response(self):
        data = ImageAd.from_dict(DictUtils.pick_fields(self.to_dict(), IMAGE_AD_PICKABLE_FIELDS))
        return ImageAd.from_dict(DictUtils.omit_fields(data.to_dict(), ["owner", "createdBy"]))

    def getChangedFields(self, data):
        last_value = self.to_dict();
        filtered_value = pydash.pick(data, ImageAdFields().filtered_keys('editable', True));
        new_value = DictUtils.merge_dict(filtered_value, last_value, True);
        _changed_fields = DictUtils.compare_nested_objects(last_value, new_value);
        changed_fields = ArrayUtils.filter(DictUtils.get_keys(_changed_fields), lambda x: str(x).find('.') == -1)
        return changed_fields
    
    def generateAd(self, adgroup, data=None, only_generate = False):
        if adgroup is None:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error("Adgroup not found","INVALID_REQUEST", 1))

        model = ImageAd.generate_model()
        model['accountId'] = adgroup.accountId
        model['clientCustomerId'] = adgroup.clientCustomerId
        if data is not None:
            model = DictUtils.merge_dict(data, model, True)
        # print('create ad with model', model)
        create = ImageAd().create(model, adgroup, only_generate)
        if create.status == 'error':
            return create
        values = create.data
        return ApiResponse(ResponseStatus.OK, StatusCode.status_200, values, None)
    
    def createAssetFile(self):
        file = None
        if bool(self.hash) is False:
            if bool(self.url) is True:
                try:
                    print('checking if file exists', self.url)
                    f = File(collection_name=DISPLAY_ADS_UPLOAD_COLLECTION).query(
                        query_params=cloud_firestore_query_builder(
                        cloud_firestore_query_formatter("file_url", self.url)
                    ), first=True)
                    if f is not None:
                        return ApiResponse(ResponseStatus.OK, StatusCode.status_200, f, None)
                    file = upload_image_from_url(file=self.url, upload_path='', extensions=IMAGES_EXTENSIONS.copy(),accountId=self.accountId, upload_folder=DISPLAY_ADS_UPLOAD_COLLECTION, process=False)
                    if file is not None:
                        return ApiResponse(ResponseStatus.OK, StatusCode.status_200, file, None)
                except Exception as e:
                    print(e)
                    return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"Cannot build asset file ref for {self.id} check if {self.url} is a valid image url", "INVALID_REQUEST", 1))
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"Cannot build asset file ref for {self.id}", "INVALID_REQUEST", 1))
        file = File({"id": self.hash}, collection_name=DISPLAY_ADS_UPLOAD_COLLECTION).get()
        if file is not None:
            return ApiResponse(ResponseStatus.OK, StatusCode.status_200, file, None)
        return ApiResponse(ResponseStatus.OK, StatusCode.status_204, None, None)
    def build_ad_file(self, file: File | None = None, adgroup=None):
        _file = file
        if _file is None and bool(self.hash) is True:
            print('checking if file exists', self.hash)
            _file = File(file={"id": self.hash}, collection_name=DISPLAY_ADS_UPLOAD_COLLECTION).get()
            print('file hash', _file)
        if _file is None:
            asset_create = self.createAssetFile()
            return asset_create
        return ApiResponse(ResponseStatus.OK, StatusCode.status_200, _file, None)

    def check_customer_used(self):
        adgroup = self.adgroup
        if adgroup is None:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error("Adgroup not found to check ad customer","INVALID_REQUEST", 1))
        
        if self.assetId is not None and self.assetId > 0:
            if self.clientCustomerId != adgroup.clientCustomerId:
                copy = self.to_dict().copy()
                copy['clientCustomerId'] = adgroup.clientCustomerId
                if "id" in copy:
                    del copy["id"]
                if "assetId" in copy:
                    del copy["assetId"]
                copy["name"] = f"Ad copy {uuid.uuid1()}"
                new_ad = ImageAd.from_dict(copy)
                new_ad.adgroup = adgroup
                new_ad.clientCustomerId = adgroup.clientCustomerId
                return ApiResponse(ResponseStatus.OK, StatusCode.status_200, new_ad, None)
        return ApiResponse(ResponseStatus.OK, StatusCode.status_200, self, None)
    def check_point_ad(self, file: File | None = None, adgroup: Any | None=None):
        if file is None:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"Cannot find or create asset file to use {self.id}", "INVALID_REQUEST", 1))
        if bool(self.name) is False:
            self.name = f"New Ad upload {uuid.uuid1()}"
        dim_valid = False
        if file.size > 150000:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"Invalid ad size {file.size} max 150kb", "INVALID_REQUEST", 1))
        for r in DISPLAY_ADS_REQUIREMENTS.copy():
            if file.width == r['width'] and file.height == r['height']:
                dim_valid = True
                break
        if dim_valid is False:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"Invalid ad dimension {self.width}x{self.height}", "INVALID_REQUEST", 1))
        return ApiResponse(ResponseStatus.OK, StatusCode.status_200, None, None)
        
    def checkpoint_asset_account(self, act: str | Account | None = None, state: str | None = None):
        error = {"state": state if state is not None else self.state}
        account = self.account;
        id = self.accountId
        if  bool(id) is False:
            if act is not None:
                if isinstance(act, str):
                    id = act
                else:
                    id = act.id
        if bool(id) and (account is None or bool(account) is False or bool(account.id) is False):
            account = Account(id).get()
        if account is not None:
            self.account = account
            return ApiResponse(ResponseStatus.OK, StatusCode.status_200, account, None)
        error["message"] = f"Cannot found account for asset {self.id}"
        return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(error, "INVALID_REQUEST", 1))
    def checkpoint_asset_customer(self, customer: str | ClientCustomerId | None = None, state: str | None = None):
        error = {"state": self.state if self.state is not None else state}
        clientCustomer = self.customer
        if clientCustomer is not None:
            return ApiResponse(ResponseStatus.OK, StatusCode.status_200, clientCustomer, None)
        id = self.clientCustomerId
        if id is None:
            if customer is not None:
                if isinstance(customer, str):
                    id = customer
                else:
                    id = customer.id
        if clientCustomer is None or bool(clientCustomer) is False or bool(clientCustomer.id) is False:
            if bool(id):
                clientCustomer = ClientCustomerId(id).get()
            else:
                customers = ClientCustomerId().getDefault()
                if bool(customers) is True:
                    clientCustomer = customers[0]
                    self.customer = clientCustomer
                    return ApiResponse(ResponseStatus.OK, StatusCode.status_200, clientCustomer, None)
        if clientCustomer is not None:
            self.customer = clientCustomer
            return ApiResponse(ResponseStatus.OK, StatusCode.status_200, clientCustomer, None)
        error["message"] = f"Cannot found customer {id} for asset {self.id}"
        return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(error, "INVALID_REQUEST", 1))
    
    def basic_checkpoint(self) -> 'ApiResponse':
        account_check = self.checkpoint_asset_account()
        if account_check.status == ResponseStatus.ERROR:
            return account_check
        account: Account = account_check.data
        customer_check = self.checkpoint_asset_customer()
        if customer_check.status == ResponseStatus.ERROR:
            return customer_check
        customer: ClientCustomerId = customer_check.data
        return ApiResponse(ResponseStatus.OK, StatusCode.status_200, (account, customer), None)

    def build_ad(self, file: File | None = None, fetch=False):
        adgroup = self.adgroup
        if adgroup is None:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error("Adgroup not found to check ad customer","INVALID_REQUEST", 1))
        image_copy = ImageAd.from_dict(self.to_dict().copy())
        ad_data = self
        if fetch is True and bool(self.id) is True:
            ad_data = self.get()
        if ad_data is None:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"Cannot find asset {self.id}", "INVALID_REQUEST", 1))
        ad_data.adgroup = image_copy.adgroup 
        base_check = ad_data.basic_checkpoint()
        if base_check.status == ResponseStatus.ERROR:
            return base_check
        account,customer = base_check.data
        ad_data.accountId = account.id
        ad_data.clientCustomerId = customer.id
        ad_data.owner = account.owner
        ad_data.adgroup_id_parent = adgroup.id
        _file = ad_data.build_ad_file(file)
        if _file.status == ResponseStatus.ERROR:
            return _file
        data_file: File = _file.data
        if data_file is None:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"Cannot find asset {self.id}", "INVALID_REQUEST", 1))
        ad_data.url = data_file.file_url
        ad_data.hash = data_file.id
        ad_data.width = data_file.width
        ad_data.height = data_file.height
        ad_data.fileSize = data_file.size
        ad_data.adgroup = adgroup
        ad_data.customer = customer
        ad_data.account = account
        d = ad_data.check_customer_used()
        if d.status == ResponseStatus.ERROR:
            return d
        if isinstance(d.data, ImageAd):
            ad_data = d.data
        check = ad_data.check_point_ad(_file.data, ad_data.adgroup)
        if check.status == ResponseStatus.ERROR:
            return check
        return ApiResponse(ResponseStatus.OK, StatusCode.status_200, ad_data, None)
        
    def create(self, data, adgroup, only_generate = False):
        # creation_date = created_at * 1000
        # return DateUtils.from_timestamp(creation_date).isoformat()
        # try:
        # print('campaign account', data)
        self.adgroup = adgroup
        data_create = DictUtils.dict_from_keys(data, ImageAdFields().filtered_keys('mutable', True));
        if bool(data_create) is False:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error("Invalid request","INVALID_REQUEST", 1))
        data_create = DictUtils.merge_dict({
                "accountId": adgroup.accountId,
                "owner": adgroup.owner,
                "clientCustomerId": adgroup.clientCustomerId,
                "adgroup_id_parent": adgroup.id,
                "createdBy": f"users/{adgroup.owner}"
            }, data_create,True)
        ad = ImageAd.from_dict(data_create)
        ad.adgroup = adgroup
        build = ad.build_ad()
        if build.status == ResponseStatus.ERROR:
            return build
        if isinstance(build.data, ImageAd):
            ad = build.data
        # print('data_create', data_create)
        document_id = self.collection().document().id
        # exist = self.query(ImageAd, "ad", [{"key": "name", "comp": "==", "value": data_create['name']}], True)
        # if exist is not None:
        #     return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"Ad name {data_create['name']} already exists","INVALID_REQUEST", 1))
        ad.id = document_id
        create = {**ad.to_dict(), 'createdAt': firestore.SERVER_TIMESTAMP}
        if only_generate is False or only_generate is None:
            doc_ref = self.collection().document(document_id);
            doc_ref.set(create, True);
        ad = ImageAd.from_dict(create).normalize_response()
        if ad is None:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error("Failed to create ad","INVALID_REQUEST", 1))
        return ApiResponse(ResponseStatus.OK, StatusCode.status_200, ad.normalize_response(), None);  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the
        # except Exception as e:
        #     print(e)
        #     return {"error": "Exception creating campaign"};

    def update(self, data):
        try:
            last_value = self.to_dict();
            print('data update', data)
            # print('last', last_value);
            filtered_value = pydash.pick(data, ImageAdFields().filtered_keys('editable', True));
            # print('filtered', filtered_value);
            new_value = DictUtils.merge_dict(filtered_value, last_value, True);
            _changed_fields = DictUtils.compare_nested_objects(last_value, new_value);
            changed_fields = ArrayUtils.filter(DictUtils.get_keys(_changed_fields), lambda x: str(x).find('.') == -1)
            data_update = DictUtils.dict_from_keys(filtered_value, changed_fields);
            if bool(data_update) is False:
                return ApiResponse(ResponseStatus.OK, StatusCode.status_200, data_update, None);
            # try:
            #     self.document_reference().set(data_update, merge=True)
            # except Exception as e:
            #     return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(str(e),"INVALID_REQUEST", 1))
            # return DictUtils.dict_from_keys(self.get(), changed_fields);
            # fields = changed_fields
            print('data_update ad', data_update)
            return ApiResponse(ResponseStatus.OK, StatusCode.status_200, data_update, None);
        except Exception as e:
            print(e)
            return None;

    def to_json(self, _fields=None):
        return self.to_dict()
    def remove(self, only_mark_as_removed=True):
        if self.is_removed is True:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_500, None, Error(f"Adgroup {self.id} already removed","INVALID_REQUEST", 1));
        try:
            if self.id is None:
                return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"Cannot identify Campaign with id {self.id}","INVALID_REQUEST", 1));
            if only_mark_as_removed:
                self.document_reference().set({"is_removed": True}, merge=True)
            else:
                self.document_reference().delete();
            return ApiResponse(ResponseStatus.OK, StatusCode.status_200, {"message": f"Campaign {self.id} deleted"}, None);
        except:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"An error occurated while removing authorization code with id {self.id}","INVALID_REQUEST", 1));




from dataclasses import dataclass
from typing import Any
from  ....utils.utils import DictUtils, BaseFieldsClass, BaseClass,init_class_kwargs,get_object_model_class

@dataclass
class DevicesFields(BaseFieldsClass):
    id = "id"
    name = "name"
    type = "type"
    criterionId = "criterionId"
    criterionType = "criterionType"
    isTargeted = "isTargeted"
    isExcluded = "isExcluded"
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

@dataclass(init=False)
class Devices(BaseClass):
    id: str
    name: str
    type: str
    criterionType: str
    criterionId: int
    isTargeted: bool
    isExcluded: bool
    criteria: dict
    def __init__(self, device=None, **kwargs):
        (cls_object, keys, data_args) = init_class_kwargs(self, device, None, None, None, [], **kwargs)
        for key in keys:
            setattr(self, key, cls_object[key]) 
        self.criterionType = "PLATFORM"
        self.type = "PLATFORM"
    def to_dict(self, field=None):
        return DictUtils.remove_none_values({
            "id": self.id,
            "name": self.name,
            "type": self.type,
            "criterionType": self.criterionType,
            "criterionId": self.criterionId,
            "isTargeted": self.isTargeted,
            "isExcluded": self.isExcluded,
            "criteria": self.criteria
        })
    @staticmethod
    def from_dict(obj: Any) -> 'Devices':
        cls_object, keys = get_object_model_class(obj, Devices, None);
        return Devices(cls_object)



from dataclasses import dataclass
from .....utils.utils import DictUtils
from datetime import datetime, timedelta
import os

CODE_COLLECTION = os.environ.get('CODE_COLLECTION');


def is_expired(expired_at):
    return datetime.fromisoformat(expired_at) < datetime.now()

def get_code_expire_at(expires_in):
    return datetime.now() + timedelta(seconds=expires_in)

@dataclass
class CodeFields:
    id = "id"
    code = "code"
    target = "target"
    code_type = "code_type"
    createdAt = "created_at"
    expired_at = "expired_at"
    expires_in = "expires_in"

    @staticmethod
    def keys():
        return DictUtils.get_keys(CodeFieldsProps);

    @staticmethod
    def filtered_keys(field, condition=True):
        mutable = DictUtils.filter(CodeFieldsProps, DictUtils.get_keys(CodeFieldsProps), field, condition)
        return DictUtils.get_keys(mutable);

CodeFieldsProps = {
    CodeFields.id: {
        "type": str,
        "required": True,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
    CodeFields.code: {
        "type": str,
        "required": True,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
    CodeFields.code_type: {
        "type": str,
        "required": True,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
    CodeFields.target: {
        "type": str,
        "required": True,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
    CodeFields.createdAt: {
        "type": str,
        "required": True,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
    CodeFields.expires_in: {
        "type": int,
        "required": True,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "default_value": 900,
        "pickable": True
    },
    CodeFields.expired_at: {
        "type": float,
        "required": True,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "default_value": get_code_expire_at(900).isoformat(),
        "pickable": True
    },

    
}
STANDARD_FIELDS = CodeFields.filtered_keys('pickable', True)


from .models import (
    CampaignDisplay, CampaignDisplayFields, 
    CampaignDisplayFieldsProps, filter_campaign_display_request_fields,
    CAMPAIGN_DISPLAY_PICKABLE_FIELDS, CAMPAIGN_DISPLAY_NOT_PICKABLE_FIELDS,
    Asset, AssetsFields, AssetsFieldsProps, ASSETS_PICKABLE_FIELDS
)



import uuid
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Any, Optional

from adafri.v1.base.firebase_collection import FirebaseCollectionBase

MEMBER_COLLECTION = 'workflow_members'

MEMBER_PICKABLE_FIELDS = [
    'id', 'workspaceId', 'uid', 'email', 'displayName',
    'role', 'teamIds', 'status', 'joinedAt',
]

MEMBER_MUTABLE_FIELDS = ['role', 'displayName', 'teamIds', 'status']

MEMBER_ROLES = ['owner', 'admin', 'editor', 'viewer', 'guest']


def _now() -> str:
    return datetime.now(timezone.utc).isoformat()


def _new_id() -> str:
    return str(uuid.uuid4())


class WorkflowMemberFields:
    @staticmethod
    def check_invite_fields(data: dict) -> dict:
        if not data:
            return {'error': 'Missing data'}
        if not data.get('workspaceId', '').strip():
            return {'error': 'workspaceId is required'}
        if not data.get('email', '').strip():
            return {'error': 'email is required'}
        role = data.get('role', 'viewer')
        if role not in MEMBER_ROLES:
            return {'error': f"role must be one of: {', '.join(MEMBER_ROLES)}"}
        return data

    @staticmethod
    def mutable_keys() -> list:
        return MEMBER_MUTABLE_FIELDS


@dataclass(init=False)
class WorkflowMember(FirebaseCollectionBase):
    id: str
    workspaceId: str
    uid: str
    email: str
    displayName: str
    role: str
    teamIds: list
    status: str
    joinedAt: str

    def __init__(self, data=None, **kwargs):
        super().__init__(collection_name=MEMBER_COLLECTION, **kwargs)
        d = data or {}
        self.id = d.get('id') or _new_id()
        self.workspaceId = d.get('workspaceId', '')
        self.uid = d.get('uid', '')
        self.email = d.get('email', '')
        self.displayName = d.get('displayName', '')
        self.role = d.get('role', 'viewer')
        self.teamIds = list(d.get('teamIds') or [])
        self.status = d.get('status', 'active')
        self.joinedAt = d.get('joinedAt') or _now()

    @classmethod
    def from_dict(cls, obj: Any, **kwargs) -> 'WorkflowMember':
        return cls(obj, **kwargs)

    def to_dict(self) -> dict:
        return {
            'id': self.id,
            'workspaceId': self.workspaceId,
            'uid': self.uid,
            'email': self.email,
            'displayName': self.displayName,
            'role': self.role,
            'teamIds': self.teamIds,
            'status': self.status,
            'joinedAt': self.joinedAt,
        }

    @staticmethod
    def toListDict(items: list, fields=None) -> list:
        return [item.to_dict() for item in items]

    @classmethod
    def find(cls, member_id: str) -> 'Optional[WorkflowMember]':
        inst = cls()
        doc = inst.collection().document(member_id).get()
        if not doc.exists:
            return None
        return cls({**doc.to_dict(), 'id': doc.id})

    @classmethod
    def findByUid(cls, workspace_id: str, uid: str) -> 'Optional[WorkflowMember]':
        inst = cls()
        docs = (
            inst.collection()
            .where('workspaceId', '==', workspace_id)
            .where('uid', '==', uid)
            .limit(1)
            .stream()
        )
        for d in docs:
            return cls({**d.to_dict(), 'id': d.id})
        return None

    @classmethod
    def listByWorkspace(cls, workspace_id: str) -> 'list[WorkflowMember]':
        inst = cls()
        docs = (
            inst.collection()
            .where('workspaceId', '==', workspace_id)
            .order_by('joinedAt')
            .stream()
        )
        return [cls({**d.to_dict(), 'id': d.id}) for d in docs]

    def save(self) -> 'WorkflowMember':
        self.collection().document(self.id).set(self.to_dict())
        return self

    def updateRole(self, role: str) -> 'WorkflowMember':
        if role not in MEMBER_ROLES:
            raise ValueError(f"Invalid role: {role}")
        self.collection().document(self.id).update({'role': role})
        self.role = role
        return self

    def delete(self) -> None:
        self.collection().document(self.id).delete()



from .models import (
    client,
    client_fields,
    code,
    code_fields,
    grant,
    grant_fields,
    token,
    token_fields,
)
from .models.client import OAuthClient
from .models.client_fields import ClientFieldProps, ClientFields
from .models.code import Code
from .models.code_fields import CodeFields, CodeFieldsProps
from .models.grant import (
    AuthorizationCodeGrant,
    OAuthGrant,
    OpenIDAuthorizationCodeGrant,
    OpenIDCode,
    OpenIDHybridGrant,
    OpenIDImplicitGrant,
)
from .models.grant_fields import GrantFields, GrantFieldsProps
from .models.token import (
    JwtTokenGenerator,
    JwtTokenValidator,
    OAuthToken,
    RefreshTokenGrant,
    TokenGenerator,
    TokenRevocationEndpoint,
    TokenIntrospectionEndpoint,
    TokenValidator,
)
from .models.token_fields import TokenFields, TokenFieldsProps
from .models.mfa import MFACredential
from .models.mfa_fields import MFAFields, MFAFieldsProps
from .models.login_session import LoginSession
from .models.login_session_fields import LoginSessionFields, LoginSessionFieldsProps
from .models.audit_log import AuditLog
from .models.audit_log_fields import AuditLogFields, AuditLogFieldsProps
from .models.device_code import DeviceCode
from .models.device_code_fields import DeviceCodeFields, DeviceCodeFieldsProps

# from .key_handler import public_jwk
from .server import (
    authorization_server,
    config_oauth,
    config_oidc_oauth,
    guard_account,
    guards_oidc,
    is_admin,
    oidc_authorization_server,
    require_oauth,
    require_oidc_oauth,
    JWKSManager,
    jwks_manager,
    MFAService,
    mfa_service,
    log_event,
)
from .token_debug import check_token, is_existing_token



import hashlib
import secrets
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Any

from .....utils import DictUtils, Crypto, get_object_model_class, init_class_kwargs
from ....base.firebase_collection import FirebaseCollectionBase, getTimestamp
from .login_session_fields import LoginSessionFields, LoginSessionFieldsProps, STANDARD_FIELDS, LOGIN_SESSION_COLLECTION
from .....utils.response import ApiResponse, Error, ResponseStatus, StatusCode


@dataclass(init=False)
class LoginSession(FirebaseCollectionBase):
    id: str
    user_uid: str
    session_token: str
    device_name: str
    device_type: str
    ip_address: str
    user_agent: str
    is_active: bool
    mfa_verified: bool
    created_at: str
    last_seen_at: str
    revoked_at: str

    def __init__(self, session=None, **kwargs):
        if type(session) is str:
            session = {"id": session}
        (cls_object, keys, data_args) = init_class_kwargs(self, session, STANDARD_FIELDS, LoginSessionFieldsProps, LOGIN_SESSION_COLLECTION, ['id'], **kwargs)
        super().__init__(**data_args)
        for key in keys:
            setattr(self, key, cls_object[key])

    @staticmethod
    def generate_model(_key_="default_value"):
        model = {}
        props = LoginSessionFieldsProps
        for k in DictUtils.get_keys(props):
            model[k] = props[k][_key_]
        return model

    @staticmethod
    def from_dict(session: Any = None, db=None, collection_name=None) -> 'LoginSession':
        cls_object, keys = get_object_model_class(session, LoginSession, LoginSessionFieldsProps)
        return LoginSession(cls_object, db=db, collection_name=collection_name)

    def query(self, query_params: list, first=False, limit=None):
        try:
            result = []
            query_result = self.custom_query(query_params, first=first, limit=limit)
            if bool(query_result):
                if first:
                    return LoginSession.from_dict(session=query_result, db=self.db, collection_name=self.collection_name)
                else:
                    for doc in query_result:
                        result.append(LoginSession.from_dict(session=doc, db=self.db, collection_name=self.collection_name))
                    return result
            return None if first else []
        except Exception as e:
            print('Exception while querying LoginSession', e)
            return None if first else []

    def to_dict(self):
        return {
            "id": self.id,
            "user_uid": self.user_uid,
            "session_token": self.session_token,
            "device_name": self.device_name,
            "device_type": self.device_type,
            "ip_address": self.ip_address,
            "user_agent": self.user_agent,
            "is_active": self.is_active,
            "mfa_verified": self.mfa_verified,
            "created_at": self.created_at,
            "last_seen_at": self.last_seen_at,
            "revoked_at": self.revoked_at,
        }

    def to_json(self, _fields=None):
        return self.to_dict()

    def get_by_token(self, raw_token: str) -> 'LoginSession':
        hashed = hashlib.sha256(raw_token.encode()).hexdigest()
        return self.query([{"key": LoginSessionFields.session_token, "comp": "==", "value": hashed}], first=True)

    def list_by_user(self, uid: str):
        return self.query([
            {"key": LoginSessionFields.user_uid, "comp": "==", "value": uid},
            {"key": LoginSessionFields.is_active, "comp": "==", "value": True}
        ], first=False)

    def revoke(self, session_id: str):
        now = datetime.now(timezone.utc).isoformat()
        doc_ref = self.document_reference(session_id)
        doc_ref.set({"is_active": False, "revoked_at": now}, merge=True)

    def revoke_all(self, uid: str):
        sessions = self.list_by_user(uid)
        now = datetime.now(timezone.utc).isoformat()
        for s in sessions:
            self.document_reference(s.id).set({"is_active": False, "revoked_at": now}, merge=True)

    def touch(self, session_id: str):
        now = datetime.now(timezone.utc).isoformat()
        self.document_reference(session_id).set({"last_seen_at": now}, merge=True)

    def create(self, user_uid: str, ip_address: str = "", user_agent: str = "",
               device_name: str = "", device_type: str = "", mfa_verified: bool = False) -> tuple:
        raw_token = secrets.token_hex(32)
        hashed_token = hashlib.sha256(raw_token.encode()).hexdigest()
        now = datetime.now(timezone.utc).isoformat()
        model = {
            "user_uid": user_uid,
            "session_token": hashed_token,
            "device_name": device_name,
            "device_type": device_type,
            "ip_address": ip_address,
            "user_agent": user_agent,
            "is_active": True,
            "mfa_verified": mfa_verified,
            "created_at": now,
            "last_seen_at": now,
            "revoked_at": "",
        }
        obj = LoginSession.from_dict(model)
        obj.id = Crypto().generate_id(user_uid + "~session~" + raw_token[:8])
        doc_ref = LoginSession(obj.to_json()).document_reference()
        doc_ref.set({**obj.to_json(), "createdAt": getTimestamp()}, merge=True)
        return obj, raw_token



from .models import file, file_fields
from .models.file import File, filter_file_request_fields
from .models.file_fields import FileFieldProps, FileFields
from .utils import (
    crop_image,
    crop_image_from_url,
    estimate_cropped_file_size,
    estimate_pre_crop_file_size,
    get_compression_estimate,
    get_file_info_from_url,
    get_file_size_from_url,
    upload_image_from_file,
    upload_image_from_string,
    upload_image_from_url,
)

IMAGES_EXTENSIONS = ['.png', '.jpg', '.jpeg', '.gif']



import uuid
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Any

from adafri.v1.base.firebase_collection import FirebaseCollectionBase

DISCUSSION_COLLECTION = 'workflow_discussion'

DISCUSSION_PICKABLE_FIELDS = [
    'id', 'workspaceId', 'authorUid', 'authorName',
    'content', 'type', 'createdAt',
]

DISCUSSION_TYPES = ['comment', 'status_change', 'approval']


def _now() -> str:
    return datetime.now(timezone.utc).isoformat()


def _new_id() -> str:
    return str(uuid.uuid4())


@dataclass(init=False)
class WorkflowDiscussionMessage(FirebaseCollectionBase):
    id: str
    workspaceId: str
    authorUid: str
    authorName: str
    content: str
    type: str
    createdAt: str

    def __init__(self, data=None, **kwargs):
        super().__init__(collection_name=DISCUSSION_COLLECTION, **kwargs)
        d = data or {}
        self.id = d.get('id') or _new_id()
        self.workspaceId = d.get('workspaceId', '')
        self.authorUid = d.get('authorUid', '')
        self.authorName = d.get('authorName', '')
        self.content = d.get('content', '')
        self.type = d.get('type', 'comment')
        self.createdAt = d.get('createdAt') or _now()

    @classmethod
    def from_dict(cls, obj: Any, **kwargs) -> 'WorkflowDiscussionMessage':
        return cls(obj, **kwargs)

    def to_dict(self) -> dict:
        return {
            'id': self.id,
            'workspaceId': self.workspaceId,
            'authorUid': self.authorUid,
            'authorName': self.authorName,
            'content': self.content,
            'type': self.type,
            'createdAt': self.createdAt,
        }

    @staticmethod
    def toListDict(items: list, fields=None) -> list:
        return [item.to_dict() for item in items]

    @classmethod
    def listByWorkspace(cls, workspace_id: str) -> 'list[WorkflowDiscussionMessage]':
        inst = cls()
        docs = (
            inst.collection()
            .where('workspaceId', '==', workspace_id)
            .order_by('createdAt')
            .stream()
        )
        return [cls({**d.to_dict(), 'id': d.id}) for d in docs]

    def save(self) -> 'WorkflowDiscussionMessage':
        self.collection().document(self.id).set(self.to_dict())
        return self



from .display_fields import (CampaignDisplayFields, CampaignDisplayFieldsProps, CAMPAIGN_DISPLAY_PICKABLE_FIELDS, CAMPAIGN_DISPLAY_NOT_PICKABLE_FIELDS)
from .display import (CampaignDisplay, filter_campaign_display_request_fields)
from .asset import (AssetsFields, AssetsFieldsProps, ASSETS_PICKABLE_FIELDS)
from .asset import (Asset, filter_asset_request_fields)


import uuid
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Any, Optional

from adafri.v1.base.firebase_collection import FirebaseCollectionBase

WORKSPACE_COLLECTION = 'workflow_workspaces'

WORKSPACE_PICKABLE_FIELDS = [
    'id', 'accountId', 'campaignIds', 'name', 'description',
    'logo', 'color', 'industry', 'website',
    'status', 'assignees', 'dueDate', 'tags',
    'ownerUid', 'adminUids',
    'createdAt', 'updatedAt', 'createdBy',
]

WORKSPACE_MUTABLE_FIELDS = [
    'name', 'description', 'logo', 'color', 'industry', 'website',
    'status', 'assignees', 'dueDate', 'tags', 'campaignIds',
    'adminUids',
]

WORKSPACE_STATUS = ['draft', 'in_review', 'approved', 'live', 'completed']


def _now() -> str:
    return datetime.now(timezone.utc).isoformat()


def _new_id() -> str:
    return str(uuid.uuid4())


class WorkflowWorkspaceFields:
    @staticmethod
    def check_create_fields(data: dict) -> dict:
        if not data:
            return {'error': 'Missing data'}
        if not data.get('name', '').strip():
            return {'error': 'name is required'}
        return data

    @staticmethod
    def mutable_keys() -> list:
        return WORKSPACE_MUTABLE_FIELDS


@dataclass(init=False)
class WorkflowWorkspace(FirebaseCollectionBase):
    id: str
    accountId: str
    campaignIds: list
    name: str
    description: str
    logo: Optional[str]
    color: Optional[str]
    industry: Optional[str]
    website: Optional[str]
    status: str
    assignees: list
    dueDate: Optional[str]
    tags: list
    ownerUid: str
    adminUids: list
    createdAt: str
    updatedAt: str
    createdBy: str

    def __init__(self, data=None, **kwargs):
        super().__init__(collection_name=WORKSPACE_COLLECTION, **kwargs)
        d = data or {}
        self.id = d.get('id') or _new_id()
        self.accountId = d.get('accountId', '')
        # Migrate legacy single-campaign workspaces to the list format
        legacy = d.get('campaignId')
        self.campaignIds = list(d.get('campaignIds') or ([legacy] if legacy else []))
        self.name = d.get('name', '')
        self.description = d.get('description', '')
        self.logo = d.get('logo')
        self.color = d.get('color')
        self.industry = d.get('industry')
        self.website = d.get('website')
        self.status = d.get('status', 'draft')
        self.assignees = list(d.get('assignees') or [])
        self.dueDate = d.get('dueDate')
        self.tags = list(d.get('tags') or [])
        self.ownerUid = d.get('ownerUid', '')
        self.adminUids = list(d.get('adminUids') or [])
        self.createdAt = d.get('createdAt') or _now()
        self.updatedAt = d.get('updatedAt') or _now()
        self.createdBy = d.get('createdBy', '')

    @classmethod
    def from_dict(cls, obj: Any, **kwargs) -> 'WorkflowWorkspace':
        return cls(obj, **kwargs)

    def to_dict(self) -> dict:
        return {
            'id': self.id,
            'accountId': self.accountId,
            'campaignIds': self.campaignIds,
            'name': self.name,
            'description': self.description,
            'logo': self.logo,
            'color': self.color,
            'industry': self.industry,
            'website': self.website,
            'status': self.status,
            'assignees': self.assignees,
            'dueDate': self.dueDate,
            'tags': self.tags,
            'ownerUid': self.ownerUid,
            'adminUids': self.adminUids,
            'createdAt': self.createdAt,
            'updatedAt': self.updatedAt,
            'createdBy': self.createdBy,
        }

    @staticmethod
    def toListDict(items: list, fields=None) -> list:
        return [item.to_dict() for item in items]

    @classmethod
    def find(cls, workspace_id: str, account_id: str = None) -> 'Optional[WorkflowWorkspace]':
        inst = cls()
        doc = inst.collection().document(workspace_id).get()
        if not doc.exists:
            return None
        data = {**doc.to_dict(), 'id': doc.id}
        if account_id and data.get('accountId') != account_id:
            return None
        return cls(data)

    @classmethod
    def listByAccount(cls, account_id: str) -> 'list[WorkflowWorkspace]':
        inst = cls()
        docs = inst.collection().where('accountId', '==', account_id).stream()
        return [cls({**d.to_dict(), 'id': d.id}) for d in docs]

    @classmethod
    def listByCampaign(cls, campaign_id: str, account_id: str) -> 'list[WorkflowWorkspace]':
        inst = cls()
        docs = (
            inst.collection()
            .where('accountId', '==', account_id)
            .where('campaignIds', 'array_contains', campaign_id)
            .stream()
        )
        return [cls({**d.to_dict(), 'id': d.id}) for d in docs]

    def save(self) -> 'WorkflowWorkspace':
        self.collection().document(self.id).set(self.to_dict())
        return self

    def update(self, data: dict) -> 'WorkflowWorkspace':
        filtered = {k: v for k, v in data.items() if k in WORKSPACE_MUTABLE_FIELDS}
        filtered['updatedAt'] = _now()
        self.collection().document(self.id).update(filtered)
        for k, v in filtered.items():
            setattr(self, k, v)
        return self

    def delete(self) -> None:
        self.collection().document(self.id).delete()



import os
from dataclasses import dataclass

from google.cloud.firestore_v1.document import Timestamp

from ....models.adgroups.ads.base_ad_fields import BaseAdFields, BaseAdFieldsProps

RESPONSIVE_DISPLAY_ADS_COLLECTION = os.environ.get('GADS_RESPONSIVE_DISPLAY_AD_COLLECTION');


@dataclass
class NativeAdToPublishFields(BaseAdFields):
    id = "id"
    titles = "titles"
    descriptions = "descriptions"
    brand = "brand"
    longHeadline = "longHeadline"
    marketingImages = "marketingImages"
    squareMarketingImages = "squareMarketingImages"
    logoImages = "logoImages"
    landscapeLogoImages = "landscapeLogoImages"
    videosAssets = "videosAssets"
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

NativeAdToPublishFieldsProps = {
    NativeAdToPublishFields.id: {
        "internal": True,
        "type": str,
        "required": True,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    NativeAdToPublishFields.brand: {
        "internal": True,
        "type": str,
        "required": True,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    NativeAdToPublishFields.titles: {
        "internal": True,
        "type": list,
        "required": True,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    NativeAdToPublishFields.descriptions: {
        "internal": True,
        "type": list,
        "required": True,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    NativeAdToPublishFields.videosAssets: {
        "internal": True,
        "type": list,
        "required": True,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    NativeAdToPublishFields.marketingImages: {
        "internal": True,
        "type": list,
        "required": True,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    NativeAdToPublishFields.squareMarketingImages: {
        "internal": True,
        "type": list,
        "required": True,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    NativeAdToPublishFields.logoImages: {
        "internal": True,
        "type": list,
        "required": True,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    NativeAdToPublishFields.landscapeLogoImages: {
        "internal": True,
        "type": list,
        "required": True,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    NativeAdToPublishFields.longHeadline: {
        "internal": True,
        "type": [str, dict],
        "required": True,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
}
@dataclass
class ResponsiveDisplayAdBaseFields(BaseAdFields):
    headlines = "headlines"
    descriptions = "descriptions"
    longHeadline = "longHeadline"
    businessName = "businessName"
    marketingImages = "marketingImages"
    squareMarketingImages = "squareMarketingImages"
    logoImages = "logoImages"
    squareLogoImages = "squareLogoImages"
    youtubeVideos = "youtubeVideos"
    resourceName = "resourceName"
    adData = "adData"
    displayUrl = "displayUrl"
    finalAppUrls = "finalAppUrls"
    finalMobileUrls = "finalMobileUrls"
    finalUrls = "finalUrls"
    is_published = "is_published"
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
@dataclass
class ResponsiveDisplayAdFields(ResponsiveDisplayAdBaseFields):
    responsive_display_ad = "responsive_display_ad"
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.class_fields_props = ResponsiveDisplayAdFieldsProps



ResponsiveDisplayAdFieldsProps = {
    **BaseAdFieldsProps,
    ResponsiveDisplayAdFields.businessName: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": "Adafri"
    },
    ResponsiveDisplayAdFields.headlines: {
        "internal": True,
        "type": list,
        "required": True,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    ResponsiveDisplayAdFields.longHeadline: {
        "internal": True,
        "type": [str, dict],
        "required": True,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    ResponsiveDisplayAdFields.descriptions: {
        "internal": True,
        "type": list,
        "required": True,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    ResponsiveDisplayAdFields.marketingImages: {
        "internal": True,
        "type": list,
        "required": True,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    ResponsiveDisplayAdFields.squareMarketingImages: {
        "internal": True,
        "type": list,
        "required": True,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    ResponsiveDisplayAdFields.logoImages: {
        "internal": True,
        "type": list,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    ResponsiveDisplayAdFields.squareLogoImages: {
        "internal": True,
        "type": list,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    ResponsiveDisplayAdFields.youtubeVideos: {
        "internal": True,
        "type": list,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    ResponsiveDisplayAdFields.adData: {
        "internal": True,
        "type": dict,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    ResponsiveDisplayAdFields.displayUrl: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    ResponsiveDisplayAdFields.finalAppUrls: {
        "internal": True,
        "type": list,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": []
    },
    ResponsiveDisplayAdFields.finalMobileUrls: {
        "internal": True,
        "type": list,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": []
    },
    ResponsiveDisplayAdFields.finalUrls: {
        "internal": True,
        "type": list,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": []
    },
    ResponsiveDisplayAdFields.responsive_display_ad: {
        "type": dict,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    ResponsiveDisplayAdFields.is_published: {
        "type": dict,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": False
    }
}

STANDARD_FIELDS = ResponsiveDisplayAdFields(fields_props=ResponsiveDisplayAdFieldsProps).filtered_keys('pickable', True)
RESPONSIVE_DISPLAY_AD_PICKABLE_FIELDS = ResponsiveDisplayAdFields(fields_props=ResponsiveDisplayAdFieldsProps).filtered_keys('pickable', True)



import base64
import hashlib
import json
import os
import random
import re
from datetime import date, datetime, time, timedelta
from typing import Any, Iterable, List, Union
from urllib.parse import parse_qs, urlparse

import pydash
from cryptography.fernet import Fernet
from google.api_core.datetime_helpers import DatetimeWithNanoseconds
from google.cloud.firestore_v1.document import Timestamp

from .response import ApiResponse, Error, ResponseStatus, StatusCode

hash = hashlib.sha1(str(os.getenv('CRYPTO_KEY')).encode())
ENCRYPTION_KEY = base64.b64encode(hash.hexdigest()[:32].encode()).decode();
from urllib.parse import unquote

camel_pat = re.compile(r'([A-Z])')
under_pat = re.compile(r'_([a-z])')


def integer(data, default=None, required=False):
    """
    Convert input data to an integer if possible.

    Parameters:
    data (Any): The input data to be converted to an integer.
    default (int, optional): The default integer to return if data is None and default is an integer. Defaults to None.
    required (bool, optional): If True, the function will not return a converted integer if data is None. Defaults to False.

    Returns:
    int or None: The converted integer if data can be converted, the default integer if specified and applicable, or None if conversion is not possible.
    """
    if isinstance(data, int):
        return data;
    if isinstance(data, str):
        try:
            return int(data);
        except Exception:
            if default is not None and isinstance(default, int) is True:
                return default
    if required is True:
        return None
    return data;
def boolean(data, required=False):
    """
    Convert input data to a boolean if possible.

    Parameters:
    data (Any): The input data to be converted to a boolean.
    required (bool, optional): If True, the function will return None if data is not a boolean. Defaults to False.

    Returns:
    bool or None: The boolean value if conversion is possible, or None if conversion is not possible and required is True.
    """

    if isinstance(data, bool):
        return data;
    if required is True:
        return None
    if type(data) is str:
        if data.lower() == 'true':
            return True;
        if data.lower() == 'false':
            return False;
    return None;

def string(data, default=None, required=False):
    """
    Convert input data to a string if possible.

    Parameters:
    data (Any): The input data to be converted to a string.
    default (str, optional): The default string to return if data is None and default is a string. Defaults to None.
    required (bool, optional): If True, the function will not return a converted string if data is None. Defaults to False.

    Returns:
    str or None: The converted string if data can be converted, the default string if specified and applicable, or None if conversion is not possible.
    """

    
    if isinstance(data, str) is True:
        return str(data);
    if bool(required) is False:
        if isinstance(data, int) is True or isinstance(data, float) is True:
            return str(data);
        if default is not None and data is None and isinstance(default, str) is True:
            return default;
        return None
    return None;

class Object:
    def __init__(self, **kwargs):
        if kwargs is not None:
            for k, v in kwargs.items():
                setattr(self, k, v);

def split_by_crlf(s):
    return [v for v in s.splitlines() if v]

def split_by_comma(s):
    sp = s.split(',')
    if len(sp) > 0:
        splitted = [];
        for split in sp:
            splitted.append(str(split).lstrip())
        return splitted;
    return []


def read_file(_filename=None, isJson=False):
    filename = _filename
    fp = filename
    if filename is None:
        return None
    if os.path.isabs(filename) is False:
        fp = os.getcwd() + "/" + filename
    content = open(fp, 'r').read()
    try:
        if isJson is True:
            return json.loads(content)
        else:
            return content
    except Exception as e:
        return None

def isBase64(data):
    try:
        decoded = decode_base64(data)
        return decoded != data
    except Exception:
            return False

def encode_if_not_base64(data) -> 'str | None':
    try:
        if isBase64(data):
            return data
        return encode_base64(data)
    except Exception as e:
        return None;
 
def decode_if_base64(data) -> 'str | None':
    try:
        if isBase64(data):
            return decode_base64(data)
        return data;
    except Exception as e:
        print('base 64 decode error', e)
        return None

def generate_random_code():
    start = random.randint(2000, 5000)
    start1 = random.randint(6000, 9000)
    end = random.randint(start, random.randint(start, start1))
    return random.randint(start, end)

def format_query_filter(key: str, value: any, comparator: str):
    """
        key A key existing in document data
        @{value} The value used to compare
        comparator The method used to compare (==, in...)
    """
    query = {};
    query["key"] = key
    query["value"] = value
    query['comp'] = comparator
    return query

def camel_to_underscore(name):
    return camel_pat.sub(lambda x: '_' + x.group(1).lower(), name)

def underscore_to_camel(name):
    return under_pat.sub(lambda x: x.group(1).upper(), name)

def convert_to_camelcase(obj):
    data = {}
    for key in pydash.keys(obj):
        data[underscore_to_camel(key)] = obj[key];
    return data;
def convert_to_underscore(obj):
    data = {}
    for key in pydash.keys(obj):
        data[camel_to_underscore(key)] = obj[key];
    return data;

def add_param_to_url(input_url,  params):
    import urllib.parse as urlparse
    from urllib.parse import urlencode
    url_parts = list(urlparse.urlparse(input_url))
    query = dict(urlparse.parse_qsl(url_parts[4]))
    query.update(params)
    url_parts[4] = urlencode(query)
    return urlparse.urlunparse(url_parts)

def get_url_params(url):
    try:
        parse_result = urlparse(url)
        query_params_dict = parse_qs(parse_result.query);
        query_params = {};
        for key in pydash.keys(query_params_dict):
            query_params[key] = ' '.join(map(str, query_params_dict[key]));
        return query_params;
    except:
        return None;


def encode_base64(data: str):
    string_bytes = data.encode("ascii")
    base64_bytes = base64.b64encode(string_bytes)
    base64_string = base64_bytes.decode("ascii")
    return base64_string;

def decode_base64(data_base64: str):
    base64_bytes = data_base64.encode("ascii")
    string_bytes = base64.b64decode(base64_bytes)
    decoded_string = string_bytes.decode("ascii")
    return unquote(decoded_string)

class Crypto:

    
    # print(hash.hexdigest())
    fernet = Fernet(ENCRYPTION_KEY);
    def encrypt(self, message:str):
        return self.fernet.encrypt(message.encode()).decode();

    def decrypt(self, encrypted:str):
        return self.fernet.decrypt(encrypted).decode();

    def hash(self, message:str):
        return hashlib.sha256(str(message).encode()).hexdigest();
    
    def generate_id(self, message:str):
        return hashlib.md5(str(message).encode()).hexdigest();

    def generate_token(self, message:str):
        encrypted_token = self.encrypt("token~"+message);
        hash_token = hashlib.md5(str(encrypted_token).encode()).hexdigest();
        return encode_base64(hash_token)
    
    def decrypt_token(self, message:str):
        try:
            d = self.decrypt(decode_base64(message));
            token_split = d.split('~');
            token = json.loads(token_split[1])
            return token
        except:
            return None;


class RequestFields:
    data = 'data';
    fields = 'fields';

def get_request_fields(field_string, fields, default_fields):
    if type(field_string) is list:
        return field_string
    
    request_fields =[];
    load_fields = [];
    if type(field_string) is list:
        load_fields = field_string
    else:
        load_fields = ArrayUtils.join_string_to_array(field_string,',');
    # print('load_fields', load_fields)
    # print('load_fields', fields)
    hasObjective = False
    if bool(load_fields):
        for f in load_fields:
            parts = parts = f.split('.')
            attr = parts[0]
            if attr == 'objective':
                # print('attr', attr)
                # print('parts', parts)
                # print('fields', fields)
                hasObjective = True
            if attr in fields:
                request_fields.append(f)
            # if len(parts) > 1:
            #     if attr in fields:
            # else:
            #     if attr in:
            #         request_fields.append(f)

        # request_fields = ArrayUtils.pick(fields, load_fields);
    
    if bool(request_fields) is False:
        request_fields = [fields[0]]
    if default_fields is not None:
        for f in default_fields:
            if f not in request_fields:
                request_fields.append(f);
    
    response = [];
    for f in request_fields:
        response.append(str(f).strip())
    
    return response;

import inspect


def get_class_properties(cls, property='__match_args__'):
        inspected = inspect.getmembers(cls, lambda a:not(inspect.isroutine(a)))
        # attr = [a for a in inspected if not(a[0].startswith('__') and a[0].endswith('__'))]
        attr = [a[1] for a in inspected if a[0]=='__match_args__']
        if(len(attr)>0):
            if type(attr[0]) is tuple:
                return list(attr[0])
        return list(attr)

def get_object_model_class(object_model, cls, fields=None, property='__match_args__'):
        obj = object_model
        if obj is None:
            obj = {};
        attributes = get_class_properties(cls, property);
        for key in attributes:
            if key not in obj:
                try:
                    obj[key] = getattr(cls, key, None)
                except Exception as err:
                    obj[key] = None

        data_object = obj.copy()
        keys = DictUtils.get_keys(data_object);
        if fields is not None:
            for key in keys:
                if key in fields:
                    # print(data_object[key])
                    if key not in data_object or data_object[key] is None or data_object[key] == 'None':
                        if 'default_value' in fields[key]:
                            data_object[key] = fields[key]['default_value']
        return data_object, keys

def init_class_kwargs(cls, obj, class_fields, class_fields_props, class_collection_name, ids_key: list[str], **kwargs):
        cls_object, keys = get_object_model_class(obj, cls, class_fields_props);
        kwargs['fields'] = class_fields
        kwargs['fields_props'] = class_fields_props

        for key in ids_key:
            documentId = None;
            if key in cls_object:
                documentId = cls_object[key]
            if documentId is not None and bool(documentId):
                # print('docId', documentId)
                kwargs['id'] = documentId;
                break;
        
        collection_name = getattr(kwargs, 'collection_name', None)
        if collection_name is None or bool(collection_name) is False:
            if class_collection_name is not None:
                kwargs['collection_name'] = class_collection_name;
        return cls_object, keys, kwargs;

class BaseClass:

    @staticmethod
    def from_dict(element, db=None, collection_name=None) -> 'Any':
        return NotImplementedError('from_json  static method must be implemented')
    def __init__(self, fields=None, fields_props=None):
        self.fields = fields;
        self.fields_props = fields_props;
    def toListDict(self, elements, fields=['id']):
        if not isinstance(elements, list):
            return None;
        result = []
        for element in elements:
            if isinstance(element, dict):
                if fields is not None and type(fields) is list and len(fields)>0:
                    result.append(DictUtils.filter_by_fields(element, fields))
                else:
                    result.append(element)
            elif getattr(element, 'to_dict', None) is not None:
                if fields is not None and type(fields) is list and len(fields)>0:
                    result.append(DictUtils.filter_by_fields(element.to_dict(), fields))
                else:
                    result.append(element.to_dict())
            elif getattr(self, 'to_json', None) is not None:
                result.append(element.to_json(fields))
        return result


    def fromListDict(self, elements, cls=None):
        if elements is None:
            return []
        result = []
        for element in elements:
            if getattr(self, 'from_dict', None) is not None:
                result.append(self.from_dict(element))
        return result
        
    def to_json(self, _fields=None):
        fields = None
        if getattr(self, 'fields', None) is not None:
            fields = self.fields
        else:
            fields = self.__dict__.items()
        if _fields is not None and type(_fields) is list and len(_fields)>0:
            fields = get_request_fields(_fields, fields, [fields[0]]);
            
        json_object, keys = get_object_model_class({}, self, getattr(self, "fields_props", None));
        if fields is None or type(fields) is not list or len(fields)==0:
            return json.loads(json.dumps(json_object, cls=JsonEncoder))
        return json.loads(json.dumps(DictUtils.filter_by_fields(json_object, fields), cls=JsonEncoder, fields=fields))
    def keep_only(self, value, keys: list[str] | None = None, except_keys: list[str] | None = None):
        if keys is None or type(keys) is not list or len(keys)==0:
            return None;
        json_object, class_keys = get_object_model_class({}, self, getattr(value, "fields_props", None));
        # print('class keys', class_keys)
        print('keys', keys)
        for key in class_keys:
            if pydash.find(keys,lambda x: x == key) is None:
                setattr(value, key, None)
            else:
                print('keep only', key)
            # if getattr(value, key) is not None:
            #     if bool(except_keys) is False or key not in except_keys:
            #         setattr(value, key, None)
        return value
    
class BaseFieldsClass:
    class_fields_props: dict
    def __init__(self, fields_props=None, **kwargs):
        self.class_fields_props = fields_props
            
    def keys(self):
        return DictUtils.get_keys(self.class_fields_props);

    def filtered_keys(self, field, condition=True):
        mutable = DictUtils.filter(self.class_fields_props, DictUtils.get_keys(self.class_fields_props), field, condition)
        return DictUtils.get_keys(mutable);

    def check_create_fields(self, request_data, fieldsProps = None):
        result = {}
        fields = fieldsProps
        if fieldsProps is None:
            fields = self.class_fields_props
        if not request_data or not isinstance(request_data, dict):
            return {"error": "Invalid request data"}
        
        for field in fields:
            if field in request_data:
                # if isinstance(request_data[field], fields[field]['type']):
                types = fields[field]['type']
                if not isinstance(types,list):
                    types = [types]
                if type(request_data[field]) in types:
                    if fields[field]['mutable']:
                        if fields[field]['required'] is False:
                            result[field] = request_data[field]
                        else:
                            if request_data[field] is not None and bool(request_data[field]) is True:
                                result[field] = request_data[field]
                            else:
                                return {"error": f"{field} is required"}
                else:
                    if fields[field]['required']:
                        return {"error": f"Invalid type for field: {field} {type(request_data[field])}"}
            else:
                if fields[field]['required']:
                    return {"error": f"Missing required field: {field}"}
        return result

    @staticmethod
    def merge_model_data(model, data):
        if not isinstance(data, dict) or not isinstance(model, dict):
            return None
        result = data
        for key in model:
            if key not in data:
                result[key] = model[key]
        return result

class JsonEncoder(json.JSONEncoder):
    def __init__(self, *args, fields=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields = fields
    def default(self, obj):
        # try:
        #     iterable = iter(obj)
        # except TypeError:
        #     pass
        # else:
        #     return list(iterable)
        # Let the base class default method raise the TypeError
        #return json.JSONEncoder.default(self, obj)
 
        attr = getattr(obj, '__dict__', None)
        to_json = getattr(obj, 'to_json', None)
        if isinstance(obj, datetime):
            return obj.isoformat()
        elif hasattr(obj, "to_dict"):
            # return {k: v for k, v in obj.__dict__.items()}
            return obj.to_dict(self.fields)
        elif hasattr(obj, "__dict__"):
            # return {k: v for k, v in obj.__dict__.items()}
            return obj.__dict__
        else:
            return None

        # if to_json is not None:
        #     # return obj.__dict__
        #     print('to json',obj.) 
        #     return obj
        # return None

class DateUtils:

    @staticmethod
    def get_current_date_time():
        return datetime.now();

    @staticmethod
    def reset_date_time(_date: datetime):
        # new_date = _date.replace(hour=0, minute=0, second=0, microsecond=0)
        return datetime.combine(_date.date(), time(0, 0, 0, 0));
    @staticmethod
    def is_past_date(date):
        try:
            now = DateUtils.reset_date_time(datetime.now())
            return now > DateUtils.reset_date_time(date);
        except Exception as e:
            print('e', e)
            return None;
    @staticmethod
    def is_superior(date: datetime, date2: datetime):
        try:
            return DateUtils.reset_date_time(date) > DateUtils.reset_date_time(date2);
        except Exception as e:
            return None;
    @staticmethod
    def is_equal(date: datetime, date2: datetime):
        try:
            return DateUtils.reset_date_time(date) == DateUtils.reset_date_time(date2);
        except Exception as e:
            return None;
    @staticmethod
    def from_iso(date_str):
        try:
            return date.fromisoformat(date_str);
        except Exception as e:
            return None;

    @staticmethod
    def from_timestamp(date_str):
        try:
            return datetime.fromtimestamp(float(date_str));
        except Exception as e:
            return None;
    @staticmethod
    def isExpired(date_str):
        try:
            return datetime.fromisoformat(date_str) < datetime.now();
        except Exception as e:
            return None;
    @staticmethod
    def convert_firestore_timestamp_to_str(firestore_timestamp: Timestamp | None, fmt: str = "%Y-%m-%d %H:%M:%S") -> str:
        """
        Convert Firestore Timestamp to a formatted string.

        Args:
            firestore_timestamp (Timestamp): The Firestore Timestamp to convert.
            fmt (str): The format for the output string (default is "%Y-%m-%d %H:%M:%S").

        Returns:
            str: The formatted string representation of the timestamp.
        """
        if firestore_timestamp is None:
            return None

        dt = None
        if isinstance(firestore_timestamp, Timestamp) is True:
            # Convert Firestore Timestamp to a datetime object
            dt = firestore_timestamp.to_datetime()
        elif isinstance(firestore_timestamp, DatetimeWithNanoseconds) is True:
            dt = DateUtils.from_timestamp(firestore_timestamp.timestamp())
        
        if dt is None:
            return firestore_timestamp
        
        # Format the datetime object to a string
        return dt.strftime(fmt)
    @staticmethod
    def getDateFromEn(date_string: str):
        slashes = len(str(date_string).split('/')) == 3
        date_string = str(date_string).replace('/', '-')
        try:
            date_format = "%Y-%m-%d"
            date_obj = datetime.strptime(date_string, date_format)
            response = {"status": "ok", "date": date_obj, "format": "en", "isSlash": slashes}
            return ApiResponse(ResponseStatus.OK, StatusCode.status_200, response, None);
        except ValueError as e:
            try:
                date_format = "%d-%m-%Y"
                date_obj = datetime.strptime(date_string, date_format)
                response = {"status": "ok", "date": date_obj, "format": "fr", "isSlash": slashes}
                return ApiResponse(ResponseStatus.OK, StatusCode.status_200, response, None);
            except:
                return ApiResponse(ResponseStatus.ERROR, StatusCode.status_500, None, Error(f"Incorrect data format {date_string}, should be YYYY-MM-DD","INVALID_REQUEST", 1));

    @staticmethod
    def format_campaign_date(data: dict, set_end_date = True):
        if not isinstance(data, dict):
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error("Invalid data","INVALID_REQUEST", 1));
        start = None
        end = None
        for key, value in data.items():
            if key =='startDate':
                start = value
            elif key == 'endDate':
                end = value
        try:
            _DATE_FORMAT_GOOGLE = "%Y%m%d"
            _DATE_FORMAT_FR = "%d/%m/%Y"
            _DATE_FORMAT_EN = "%Y-%m-%d"
            campaign = {};
            startDate = None;
            endDate = None
            startDateFormat = None;
            endDateFormat = None;
            slashes = []

            if start is not None:
                date_format = DateUtils.getDateFromEn(start);
                if date_format.status=='error':
                    return date_format;
                date_object = date_format.data
                startDate = date_object['date']
                if DateUtils.is_past_date(startDate) is True:
                    return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error("Start date cannot be in the past","INVALID_REQUEST", 1));
                startDateFormat = date_object['format']
                slashes.append(date_object['isSlash'])
                campaign['startDateFormattedGoogle'] = datetime.strftime(startDate, _DATE_FORMAT_GOOGLE);
                campaign['startDateFrench'] = datetime.strftime(startDate, _DATE_FORMAT_FR);
                campaign['startDateEnglish'] = datetime.strftime(startDate, _DATE_FORMAT_EN);
                format = None
                if startDateFormat == 'fr':
                    format = datetime.strftime(startDate, _DATE_FORMAT_FR);
                else:
                    format = datetime.strftime(startDate, _DATE_FORMAT_EN);
                if True in slashes:
                    format = format.replace('-', '/')
                campaign['startDate'] = format;

            # Optional: Set the end date.
            if startDate is not None and set_end_date is True:
                endDate = startDate + timedelta(weeks=10)
            # end_time = date + timedelta(weeks=10);
            if end is not None and bool(end) is True:
                end_date_format = DateUtils.getDateFromEn(end);
                if end_date_format.status=='error':
                    return end_date_format;
                date_object = end_date_format.data
                endDate = date_object['date'];
            
            if endDate is not None:
                if DateUtils.is_past_date(endDate) is True:
                    return ApiResponse(ResponseStatus.ERROR, StatusCode.status_500, None, Error("End date cannot be in the past","INVALID_REQUEST", 1));
                if DateUtils.is_superior(startDate, endDate) is True:
                    return ApiResponse(ResponseStatus.ERROR, StatusCode.status_500, None, Error("Start date cannot be superior to end date","INVALID_REQUEST", 1));
                if DateUtils.is_equal(startDate, endDate) is True:
                    return ApiResponse(ResponseStatus.ERROR, StatusCode.status_500, None, Error("Start date cannot be equal to end date","INVALID_REQUEST", 1));
                endDateFormat = date_object['format']
                campaign['endDateFormattedGoogle'] = datetime.strftime(endDate, _DATE_FORMAT_GOOGLE);
                campaign['endDateFrench'] = datetime.strftime(endDate, _DATE_FORMAT_FR);
                campaign['endDateEnglish'] = datetime.strftime(endDate, _DATE_FORMAT_EN);
                if 'endDate' not in data:
                    format = None
                    if startDateFormat == 'fr' or endDateFormat == 'fr':
                        format = datetime.strftime(endDate, _DATE_FORMAT_FR);
                    else:
                        format = datetime.strftime(endDate, _DATE_FORMAT_EN);
                    if True in slashes:
                        format = format.replace('-', '/')
                    campaign['endDate'] = format;

            

            return ApiResponse(ResponseStatus.OK, StatusCode.status_200, campaign, None);
        except Exception as e:
            print('exception occurated on format campaign date', e);
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"An error occurated {str(e)}","INVALID_REQUEST", 1));

class DictUtils:

    @staticmethod
    def pick_object_values(obj, keys: list[str]):
        cls_object = {};
        for attr in keys:
            cls_object[attr] = pydash.get(obj, attr, None)
        return cls_object
    @staticmethod
    def every_match(element, exceptions=[]):
        if element is None:
            return False;
    
        data = pydash.omit(element, exceptions);
        keys = pydash.keys(data);
        if bool(keys) == False:
            return False;
    
        for key in keys:
            if bool(data[key]) is False:
                print("key", key);
                print("data", data[key]);
                return False;
        return True;

    @staticmethod
    def difference(first_dict, second_dict):
        return { k : second_dict[k] for k in set(second_dict) - set(first_dict) }

    @staticmethod
    def compare_nested_objects(obj1: dict, obj2: dict, path: str = '', keys_on_dict: list[str] = []) -> dict:
        """
        Compare two nested dictionaries and return the changed fields, including lists.

        Args:
            obj1 (dict): The first dictionary to compare.
            obj2 (dict): The second dictionary to compare.
            path (str): The current path in the nested structure (used for recursion).

        Returns:
            dict: A dictionary containing the paths of changed fields and their values.
        """
        changes = {}

        # Get all keys from both dictionaries
        all_keys = set(obj1.keys()).union(set(obj2.keys()))

        for key in all_keys:
            # Create the current path
            current_path = f"{path}.{key}" if path else key

            # Check if the key exists in both dictionaries
            if key in obj1 and key in obj2:
                if isinstance(obj1[key], dict) and isinstance(obj2[key], dict):
                    # If both values are dictionaries, recurse
                    # for kk in keys_on_dict:
                    #     if kk in obj1[key] and kk in obj2[key]:
                    #         if obj1[key][kk] != obj2[key][kk]:
                    #             changes[current_path] = {'old_value': obj1[key], 'new_value': obj2[key]}
                    #             continue
                    nested_changes = DictUtils.compare_nested_objects(obj1[key], obj2[key], current_path, keys_on_dict)
                    changes.update(nested_changes)
                elif isinstance(obj1[key], list) and isinstance(obj2[key], list):
                    # If both values are lists, compare them
                    if len(obj1[key]) != len(obj2[key]):
                        changes[current_path] = {'old_value': obj1[key], 'new_value': obj2[key]}
                        continue
                    list_changes = ArrayUtils.compare_lists(obj1[key], obj2[key], current_path)
                    changes.update(list_changes)
                elif obj1[key] != obj2[key]:
                    # If values are different, record the change
                    changes[current_path] = {'old_value': obj1[key], 'new_value': obj2[key]}
            elif key in obj1:
                # Key exists only in obj1
                changes[current_path] = {'old_value': obj1[key], 'new_value': None}
            else:
                # Key exists only in obj2
                changes[current_path] = {'old_value': None, 'new_value': obj2[key]}

        return changes



    @staticmethod
    def get_changed_field(a, b):

        keys_a = pydash.keys(a);
        keys_b = pydash.keys(b);
        changed_a = [];
        changed_b = [];
        if isinstance(a, list):
            print('a', a)
            print('b', b)
            if isinstance(b, list):
                for i in range(len(a)):
                    if isinstance(a[i], dict):
                        changed = DictUtils.get_changed_field(a[i], b[i]);
                        if len(changed)>0:
                            changed_a.append(a[i]);
                            continue;
                    if a[i]!=b[i]:
                        changed_a.append(a[i]);
            else:
                changed_a.append(a);
            
            for i in range(len(b)):
                if isinstance(b[i], dict):
                    changed = DictUtils.get_changed_field(b[i], a[i]);
                    if len(changed)>0:
                        changed_b.append(i);
                        continue;
                if b[i]!=a[i]:
                    changed_b.append(i);
            changed_fields = pydash.uniq(changed_a + changed_b);
            return changed_fields;

        for k in keys_a:
            if k in b:
                if isinstance(b[k], dict):
                    changed = DictUtils.get_changed_field(b[k], a[k]);
                    if k == "phoneInfo":
                        print('changed e', changed)
                    if len(changed)>0:
                        changed_a.append(k);
                        continue;
                # print('a', a)
                # print('b', b)
                # print('k', k)
                # print('k', {a: a[k], b: b[k]})
                if a[k]!=b[k]:
                    changed_a.append(k);
            else:
                changed_a.append(k);
        
        for k in keys_b:
            if k in a:
                if b[k]!=a[k]:
                    if isinstance(a[k], dict):
                        changed = DictUtils.get_changed_field(a[k], b[k]);
                        if len(changed)>0:
                            changed_b.append(k);
                            continue;
                    changed_b.append(k);
            else:
                changed_b.append(k);
        
        changed_fields = pydash.uniq(changed_a + changed_b);
        return changed_fields;

    @staticmethod
    def dict_from_keys(a, keys):
        picked = pydash.pick(a, keys);
        for key in keys:
            if key.endswith('[0]'):
                key = key[:-3];
                if key in a and isinstance(a[key], list) and len(a[key])==0:
                    picked[key] = [];
        return picked;

    @staticmethod
    def get_keys(a):
        return pydash.keys(a);
    
    @staticmethod
    def merge_dict(current_dict, target_dict, force=False):
        final_dict = target_dict.copy()
        # print('target_dict', target_dict)
        for key in pydash.keys(current_dict):
            if force is True or key not in final_dict or final_dict[key]!=current_dict[key]:
                final_dict[key] = current_dict[key]
        return final_dict;

    @staticmethod
    def string_to_json(element):
        try:
            return json.loads(element);
        except Exception as e:
            print(e)
            return None;
    
    @staticmethod
    def json_to_string(element):
        try:
            return json.dumps(element);
        except:
            return None;

    @staticmethod
    def to_dict(element):
        d = pydash.to_dict(element);
        return d;
    @staticmethod
    def pick_fields(element, array):
        return pydash.pick(element, array);
    @staticmethod
    def omit_fields(element, array):
        return pydash.omit(element, array);

    @staticmethod
    def filter(element, keys, path, condition):
        result = {}
        for key in keys:
            if element[key][path] == condition:
                result[key] = element[key]
        return result;
    @staticmethod
    def pick(element, key, data_type, default_value=None):
            if data_type == str:
                if element is not None and key in element:
                    if bool(str(element[key])) is False:
                        if default_value is not None:
                            return default_value
                    return str(element[key]);
                else:
                    if default_value is not None:
                        return default_value
                    return '';
            if data_type == bool:
                if element is not None and key in element:
                    return bool(element[key]);
                else:
                    if default_value is not None:
                        return default_value
                    return False;
    
            if data_type == list:
                if element is not None and key in element:
                    if element[key] is None or bool(list(element[key])) is False:
                        if default_value is not None:
                            return default_value
                        else:
                            return []
                    return list(element[key]);
                else:
                    if default_value is not None:
                        return default_value
                    return [];
            if data_type == dict:
                if element is not None and key in element:
                    if element[key] is not None and bool(dict(element[key])) is False:
                        if default_value is not None:
                            return default_value
                    if element[key] is not None:
                        return dict(element[key]);
                    return None;
                else:
                    if default_value is not None:
                        return default_value
                    return None;
            if data_type == int:
                if element is not None and key in element:
                    return int(element[key]);
                else:
                    if default_value is not None:
                        return default_value
                    return 0;
            if data_type == round:
                if element is not None and key in element:
                    return round(int(str(element[key])));
                else:
                    if default_value is not None:
                        return default_value
                    return 0;
    
            if data_type == float:
                if element is not None and key in element:
                    r = float(str(element[key]))
                    print(f"{key} {r} {type(r)} {element[key]}")
                    if pydash.is_nan(r) or str(r).lower()=='nan':
                        return 0.0
                    return r;
                else:
                    if default_value is not None:
                        return default_value
                    return 0.0;
            return None;

    @staticmethod
    def filter_by_fields_(data, fields: list[str]) -> dict:
        """Filter dictionary by keeping only specified fields, including nested paths"""
        if not data or not fields:
            return {}
        if isinstance(data, list):
            print('filtering list', data, fields)
            return [DictUtils.filter_by_fields(item, fields) for item in data]
        
        if not isinstance(data, dict):
            print('filtering not dict', data, fields)
            return data;
        result = {}
        print('filtering', data, fields)
        for field in fields:
            if '.' in field:
                # Handle nested fields
                parts = field.split('.')
                value = DictUtils.remove_none_values(data)
                for part in parts:
                    value = value.get(part, {}) if isinstance(value, dict) else {}
                if value:
                    pydash.set_(result, field, value)
            else:
                # Handle top-level fields
                if field in data:
                    result[field] = data[field]
        return DictUtils.remove_none_values(result)
    @staticmethod
    def remove_none_values(obj, keep_empty=True):
        """
        Recursively remove all None values from dictionaries and lists
        
        Args:
            obj: Dict, list, or other object to clean
            keep_empty: Boolean to determine if empty dicts/lists should be kept
            
        Returns:
            Cleaned object with all None values removed
        """
        if isinstance(obj, dict):
            cleaned = {
                key: DictUtils.remove_none_values(value, keep_empty)
                for key, value in obj.items()
                if value is not None and DictUtils.remove_none_values(value, keep_empty) is not None
            }
            return cleaned if cleaned or keep_empty else None
        
        if isinstance(obj, list):
            cleaned = [
                DictUtils.remove_none_values(item, keep_empty)
                for item in obj
                if item is not None and DictUtils.remove_none_values(item, keep_empty) is not None
            ]
            return cleaned if cleaned or keep_empty else None
        
        if isinstance(obj, (str, int, float, bool)):
            return obj
            
        return None
    @staticmethod
    def filter_by_fields_(obj, path):
        """
        Extract a value from a nested object using a dot-separated path.
        
        Args:
            obj: The object to extract from (dict, list, or any nested structure)
            path: String with dot-separated path (e.g., "user.address.street")
        
        Returns:
            The value at the specified path or None if path doesn't exist
        """
        keys = path.split('.')
        result = obj
        
        try:
            for key in keys:
                if key == "marketingImages.usage":
                    print("usage key ===>", result)
                if isinstance(result, (dict, list)):
                    # Handle array indices
                    if key.isdigit() and isinstance(result, list):
                        result = result[int(key)]
                    else:
                        result = result[key]
                else:
                    return None
            return result
        except (KeyError, IndexError, TypeError):
            return None
    def filter_by_fields(obj: Any, paths: Union[List[str], str]) -> Any:
        # Convert single path to list
        if isinstance(paths, str):
            paths = [paths]
        
        # Convert paths to set for O(1) lookup
        path_set = set(paths)
        # print('path_set ==>', path_set)
            
        def should_include_value(value: Any) -> bool:
            """Check if value should be included in output"""
            return isinstance(value, (str, int, float, bool, DatetimeWithNanoseconds)) or value is None
        
        def pick_nested(obj: Any, current_path: str = '') -> Any:
            """Recursively pick values from nested structure"""
            # Handle lists
            # print('path nested ==>', current_path)
            # if current_path=='marketingImages':
                # print('usage ==>', obj)
            if isinstance(obj, list):
                picked_list = [pick_nested(item, current_path) for item in obj]
                return [item for item in picked_list if item not in (None, {}, [])]
                
            # Handle non-dict values
            if not isinstance(obj, dict):
                return obj if should_include_value(obj) else None
                
            # Handle dict
            result = {}
            for key, value in obj.items():
                # if key == 'ages' and key in path_set:
                    # print('ages ==>', value)
                if key in path_set or f"{current_path}.{key}" in path_set:
                    if isinstance(value, list):
                        if len(value) > 0:
                            if not isinstance(value[0], (dict, list)):
                                result[key] = value
                                continue
                        else:
                            result[key] = value
                            continue
                if key in path_set and type(value) is list and len(value) == 0:
                    result[key] = value
                    continue
                new_path = f"{current_path}.{key}" if current_path else key
                
                # Check if this exact path is requested
                # if new_path == 'ages.text':
                    # print('ages.text ==>', value)
                if new_path in path_set:
                    if should_include_value(value):
                        result[key] = value
                        continue
                    
                # Check if this path is a prefix of any requested path
                if any(p.startswith(new_path + '.') for p in path_set):
                    # if new_path == 'ages':
                        # print('ages.text ==>', value)
                    picked = pick_nested(value, new_path)
                    if picked not in (None, {}, []):
                        result[key] = picked
                        
            return result
        
        return pick_nested(obj)

def get_nested_value(data, path):
    keys = path.split('.')
    value = data
    for key in keys:
        if isinstance(value, dict):
            value = value.get(key)
        else:
            return None
    return value

class ArrayUtils:

    @staticmethod
    def compare_lists(list1: list, list2: list, path: str) -> dict:
        """
        Compare two lists and return the changes.

        Args:
            list1 (list): The first list to compare.
            list2 (list): The second list to compare.
            path (str): The current path in the nested structure.

        Returns:
            dict: A dictionary containing the paths of changed fields and their values.
        """
        changes = {}
        max_length = max(len(list1), len(list2))

        for index in range(max_length):
            current_path = f"{path}[{index}]"
            value1 = list1[index] if index < len(list1) else None
            value2 = list2[index] if index < len(list2) else None

            if value1 != value2:
                changes[current_path] = {'old_value': value1, 'new_value': value2}

        return changes
    @staticmethod
    def filter_array_with_another(array_to_filter, filter):
        return pydash.filter_(array_to_filter, lambda x: pydash.find(filter, lambda y: x==y) is not None);

    def filter(array, filter):
        return pydash.filter_(array, filter);

    @staticmethod
    def get_uniq_value(elements, _key, _key_=None, parse=False):
        keys = [];
        for element in elements:
            if parse is False:
                print('type', type(element[_key]))
                if type(element[_key]) == str or type(element[_key]) == dict:
                    keys.append(element[_key]);
                elif type(element[_key]) == list:
                    keys = pydash.arrays.concat(keys, element[_key])
                    print('here', keys)
            else:
                data = json.loads(JsonEncoder().encode(element))[_key]
                if type(element[_key]) == str or type(element[_key]) == dict:
                    keys.append(data)
                elif type(element[_key]) == list:
                    keys = pydash.arrays.concat(keys, data)
        if _key_ is None:
            return pydash.uniq(keys);
        else:
            return pydash.uniq_by(keys, _key_);
    @staticmethod
    def group_elements_by_key(elements, _key, parse=False):
        keys = ArrayUtils.get_uniq_value(elements, _key, parse);
        result = {};
        for key in keys:
            e = None;
            if parse is False:
                e = pydash.filter_(elements, lambda x: x[_key]==key);
            else:
                e = pydash.filter_(elements, lambda x: json.loads(JsonEncoder().encode(x))[_key]==key);
            if key not in result:
                result[key] = pydash.flatten([e]);
            else:
                result[key] = pydash.flatten(result[key].append(e));
        return result;
    
    @staticmethod
    def sum_grouped_by_key(elements, _key):
        result = []
        for key in pydash.keys(elements):
            sum = pydash.sum_by(elements[key], _key);
            result.append({'key': key, "sum": sum})
        return result;

    @staticmethod
    def string_to_array(element) -> 'list':
        """Converts a string to an array of strings, if the string is not empty."""
        try:
            return json.loads(element);
        except:
            return None;
    @staticmethod
    def to_array(element) -> 'list':
        """Turn element into an array if it is not already an array."""
        if element is None:
            return []
        try:
            if isinstance(element, list):
                return element;
            return [element];
        except:
            return [];

    
    @staticmethod
    def join_string_to_array(element, separator='') -> 'Iterable[str]':
        try:
            split = str(element).split(separator);
            if len(split) == 0:
                if len(element) > 0:
                    return [element];
                return []
            else:
                return split;
        except:
            return [];
    
    @staticmethod
    def array_include_array(element, array_check, x = None, y = None) -> 'bool':
        
        def check_filter(check_data, filter_data):
            if x is None or bool(x) is False:
                if y is None or bool(y) is False:
                    return check_data == filter_data
                else:
                    if y in check_data:
                        return check_data[y] == filter_data
                    return False
            else:
                if y is None or bool(y) is False:
                    if x in filter_data:
                        return check_data == filter_data[x]
                    return False
                else:
                    if x in filter_data and y in check_data:
                        return filter_data[x] == check_data[y]
                    return False
                
        def filter(filter_data):
            return pydash.filter_(array_check, lambda arr: check_filter(arr,filter_data))
                 
        return pydash.filter_(element,  lambda element_data: filter(element_data))

    @staticmethod
    def array_to_string(element) -> 'str':
        try:
            return json.dumps(element);
        except:
            return None;
    @staticmethod
    def pick(element, pickable_fields) -> 'list':
        return pydash.filter_(element, lambda x: pydash.includes(pickable_fields, x))
    
    @staticmethod
    def pick_iterable_childs_with_key(iterable, pickable_fields) -> 'list':
        values = [];
        
        if iterable is None or type(iterable) is not list:
            return [];
    
        for element in iterable:
            picked = pydash.pick(element,  pickable_fields);
            if picked is not None and bool(picked) is True:
                values.append(picked);
        return values
    
    @staticmethod
    def pick_iterable_values_with_key(iterable, field) -> 'list':
        values = [];
        
        if iterable is None or type(iterable) is not list:
            return [];
    
        for element in iterable:
            if field in element and bool(element[field]) is True:
                values.append(element[field]);
        return values;

    @staticmethod
    def class_list_to_list_dict(classes):
        values = [];
        for class_object in classes:
            to_json = getattr(class_object, "to_json", None)
            if callable(to_json):
                values.append(class_object.to_json());
        return values;
    @staticmethod
    def find(array, predicate):
        return pydash.find(array, predicate)


class RequestUtils:
    @staticmethod
    def get_data(request):
        print(request.data.decode())
        data_request = DictUtils.string_to_json(request.data.decode())
        print("data request",data_request)
        return data_request;


from dataclasses import dataclass
from .....utils import DictUtils
import os

DEVICE_CODE_COLLECTION = os.environ.get('DEVICE_CODE_COLLECTION', 'device_codes')


@dataclass
class DeviceCodeFields:
    id = "id"
    device_code = "device_code"
    user_code = "user_code"
    client_id = "client_id"
    scope = "scope"
    user_uid = "user_uid"
    status = "status"
    interval = "interval"
    expires_at = "expires_at"
    last_polled_at = "last_polled_at"
    created_at = "created_at"

    @staticmethod
    def keys():
        return DictUtils.get_keys(DeviceCodeFieldsProps)

    @staticmethod
    def filtered_keys(field, condition=True):
        mutable = DictUtils.filter(DeviceCodeFieldsProps, DictUtils.get_keys(DeviceCodeFieldsProps), field, condition)
        return DictUtils.get_keys(mutable)


DeviceCodeFieldsProps = {
    DeviceCodeFields.id: {
        "type": str,
        "required": True,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
    DeviceCodeFields.device_code: {
        "type": str,
        "required": True,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
    DeviceCodeFields.user_code: {
        "type": str,
        "required": True,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
    DeviceCodeFields.client_id: {
        "type": str,
        "required": True,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
    DeviceCodeFields.scope: {
        "type": str,
        "required": False,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
    DeviceCodeFields.user_uid: {
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
    DeviceCodeFields.status: {
        "type": str,
        "required": True,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "default_value": "pending",
        "pickable": True
    },
    DeviceCodeFields.interval: {
        "type": int,
        "required": True,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "default_value": 5,
        "pickable": True
    },
    DeviceCodeFields.expires_at: {
        "type": str,
        "required": True,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
    DeviceCodeFields.last_polled_at: {
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
    DeviceCodeFields.created_at: {
        "type": str,
        "required": False,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
}

STANDARD_FIELDS = DeviceCodeFields.filtered_keys('pickable', True)



from user_agents import parse as user_agents_parse
def get_request_data(_request):
    data = None;
    if _request is None:
        return data
    if 'data' in _request.json:
        data = _request.json['data']
    elif 'data' in _request.form:
        data = _request.form['data']
    return data
def get_request_operation(_request):
    operation = None;
    allowed_operations = ['create', 'update', 'remove']
    if _request is None:
        return operation
    if 'operation' in _request.json:
        operation = _request.json['operation']
    elif 'operation' in _request.form:
        operation = _request.form['operation']
    print('opt', operation)
    if operation not in allowed_operations:
        operation = None
    return operation

def to_camel_case(s: str, separator="-") -> str:
    """
    Convert a string to camel case.

    Args:
        s (str): The input string to convert.

    Returns:
        str: The camel case version of the input string.
    """
    # Split the string into words
    words = s.split(separator)
    
    # Capitalize the first letter of each word except the first one
    # and join them together
    if not words:
        return ""
    
    # Convert the first word to lowercase and the rest to title case
    camel_case_string = words[0].lower() + ''.join(word.capitalize() for word in words[1:])
    
    return camel_case_string
def parse_request_headers(request) -> dict:
    """
    Parse the request headers and return them as a dictionary.
    
    Returns:
        dict: A dictionary containing the request headers.
    """
    headers_dict = {}
    for header, value in request.headers.items():
        headers_dict[to_camel_case(header)] = value
    return headers_dict
def get_request_infos(request):
    headers = parse_request_headers()
    user_agent_string = request.headers.get('User-Agent')
    
    # Parse the user agent string
    user_agent = user_agents_parse(user_agent_string)
    # user_agent.browse
    # Create a dictionary with relevant user agent information
    user_agent_info = {
        **headers,
        'browser': {
            'family': user_agent.browser.family,
            'version': user_agent.browser.version_string,
        },
        'os': {
            'family': user_agent.os.family,
            'version': user_agent.os.version_string,
        },
        'device': {
            'family': user_agent.device.family,
            'is_mobile': user_agent.is_mobile,
            'is_tablet': user_agent.is_tablet,
            'is_pc': user_agent.is_pc,
        },
        'user_agent_string': user_agent_string
    }
    return user_agent_info


NATIVE_ADS_REQUIREMENTS = [
    # Marketing Images
    {
        "type": "marketing_image",
        "name": "Marketing Image 1.91:1",
        "minWidth": 600,
        "minHeight": 314,
        # "maxWidth": 600,
        # "maxHeight": 314,
        "maxWidth": 1200,
        "maxHeight": 628,
        "ratio": "1.91:1",
        "maxFileSize": 5120,  # 5MB in KB
        # "maxFileSize": 0.5,  # 1MB in KB
        "formats": ["JPG", "PNG"],
        "description": "Landscape marketing image for responsive display ads"
    },    
    # Square Marketing Images
    {
        "type": "square_marketing_image",
        "name": "Square Marketing Image",
        "minWidth": 300,
        "minHeight": 300,
        # "maxWidth": 600,
        # "maxHeight": 600,
        "maxWidth": 1200,
        "maxHeight": 1200,
        "ratio": "1:1",
        "maxFileSize": 5120,
        # "maxFileSize": 0.5,  # 1MB in KB
        "formats": ["JPG", "PNG"],
        "description": "Square marketing image (same as marketing image 1:1)"
    },
    # Logos
    {
        "type": "logo",
        "name": "Logo 1:1",
        "minWidth": 128,
        "minHeight": 128,
        "maxWidth": 512,
        "maxHeight": 512,
        "ratio": "1:1",
        "maxFileSize": 5120,
        # "maxFileSize": 0.5,
        "formats": ["JPG", "PNG"],
        "description": "Square logo for responsive display ads"
    },
    {
        "type": "logo_landscape",
        "name": "Logo 4:1",
        "minWidth": 512,
        "minHeight": 128,
        "maxWidth": 2048,
        "maxHeight": 512,
        "ratio": "4:1",
        "maxFileSize": 5120,
        # "maxFileSize": 0.5,
        "formats": ["JPG", "PNG"],
        "description": "Landscape logo for responsive display ads"
    }
]

# MARKETING_IMAGES_RATIO = [
#     # Marketing Images
#     {
#         "type": "marketing_image",
#         "name": "Marketing Image 1.91:1",
#         "minWidth": 600,
#         "minHeight": 314,
#         "maxWidth": 1200,
#         "maxHeight": 628,
#         "ratio": "1.91:1",
#         "maxFileSize": 5120,  # 5MB in KB
#         "formats": ["JPG", "PNG"],
#         "description": "Landscape marketing image for responsive display ads"
#     },
#     {
#         "type": "marketing_image", 
#         "name": "Marketing Image 1:1",
#         "minWidth": 300,
#         "minHeight": 300,
#         "maxWidth": 1200,
#         "maxHeight": 1200,
#         "ratio": "1:1",
#         "maxFileSize": 5120,
#         "formats": ["JPG", "PNG"],
#         "description": "Square marketing image for responsive display ads"
#     },
#     {
#         "type": "marketing_image",
#         "name": "Marketing Image 4:3",
#         "minWidth": 600,
#         "minHeight": 450,
#         "maxWidth": 1200,
#         "maxHeight": 900,
#         "ratio": "4:3",
#         "maxFileSize": 5120,
#         "formats": ["JPG", "PNG"],
#         "description": "Standard marketing image for responsive display ads"
#     },
#     {
#         "type": "marketing_image",
#         "name": "Marketing Image 16:9",
#         "minWidth": 600,
#         "minHeight": 338,
#         "maxWidth": 1200,
#         "maxHeight": 675,
#         "ratio": "16:9",
#         "maxFileSize": 5120,
#         "formats": ["JPG", "PNG"],
#         "description": "Widescreen marketing image for responsive display ads"
#     },
    
#     # Square Marketing Images
#     {
#         "type": "square_marketing_image",
#         "name": "Square Marketing Image",
#         "minWidth": 300,
#         "minHeight": 300,
#         "maxWidth": 1200,
#         "maxHeight": 1200,
#         "ratio": "1:1",
#         "maxFileSize": 5120,
#         "formats": ["JPG", "PNG"],
#         "description": "Square marketing image (same as marketing image 1:1)"
#     },
    
#     # Logos
#     {
#         "type": "logo",
#         "name": "Logo 1:1",
#         "minWidth": 128,
#         "minHeight": 128,
#         "maxWidth": 512,
#         "maxHeight": 512,
#         "ratio": "1:1",
#         "maxFileSize": 5120,
#         "formats": ["JPG", "PNG"],
#         "description": "Square logo for responsive display ads"
#     },
#     {
#         "type": "logo",
#         "name": "Logo 4:1",
#         "minWidth": 512,
#         "minHeight": 128,
#         "maxWidth": 2048,
#         "maxHeight": 512,
#         "ratio": "4:1",
#         "maxFileSize": 5120,
#         "formats": ["JPG", "PNG"],
#         "description": "Landscape logo for responsive display ads"
#     },
    
#     # Additional Marketing Image Ratios (Optional)
#     {
#         "type": "marketing_image",
#         "name": "Marketing Image 3:2",
#         "minWidth": 600,
#         "minHeight": 400,
#         "maxWidth": 1200,
#         "maxHeight": 800,
#         "ratio": "3:2",
#         "maxFileSize": 5120,
#         "formats": ["JPG", "PNG"],
#         "description": "Classic marketing image ratio"
#     },
#     {
#         "type": "marketing_image",
#         "name": "Marketing Image 2:1",
#         "minWidth": 600,
#         "minHeight": 300,
#         "maxWidth": 1200,
#         "maxHeight": 600,
#         "ratio": "2:1",
#         "maxFileSize": 5120,
#         "formats": ["JPG", "PNG"],
#         "description": "Wide marketing image ratio"
#     },
#     {
#         "type": "marketing_image",
#         "name": "Marketing Image 5:4",
#         "minWidth": 600,
#         "minHeight": 480,
#         "maxWidth": 1200,
#         "maxHeight": 960,
#         "ratio": "5:4",
#         "maxFileSize": 5120,
#         "formats": ["JPG", "PNG"],
#         "description": "Portrait marketing image ratio"
#     }
# ]


# from .display import (
#     CampaignDisplay, CampaignDisplayFields, CampaignDisplayFieldsProps,
#     AdGroupDisplay, AdGroupDisplayFields, AdGroupDisplayFieldsProps,
#     filter_campaign_display_request_fields, filter_adgroup_display_request_fields,
#     ADGROUP_DISPLAY_PICKABLE_FIELDS, CAMPAIGN_DISPLAY_PICKABLE_FIELDS, CAMPAIGN_DISPLAY_NOT_PICKABLE_FIELDS
#     )
from .models import ClientCustomerId
from .models.user_interest import UserInterest
from .utils import search_youtube_channel_by_id, search_youtube_video_by_id
